import enum
import logging
import math
from functools import cached_property, partial
from pathlib import Path
from typing import Any, Literal, Mapping, Optional, Union, cast

import torch
from lightning import LightningDataModule, LightningModule
from lightning.fabric.loggers.csv_logs import CSVLogger
from lightning.fabric.utilities.types import LRScheduler
from lightning.pytorch.utilities.types import OptimizerLRScheduler
from lightning_utilities.core.apply_func import apply_to_collection
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from torch.utils.data import DataLoader

from lightning_cv.callbacks.base import Callback, CallbackT
from lightning_cv.callbacks.checkpoint import ModelCheckpoint
from lightning_cv.callbacks.logger import LogFlusher
from lightning_cv.callbacks.mode import EvalMode, TrainMode
from lightning_cv.callbacks.progress import TqdmProgressBar
from lightning_cv.callbacks.summary import ModelSummary
from lightning_cv.callbacks.utils import standardize_callbacks
from lightning_cv.config import CrossValidationTrainerConfig, FoldState
from lightning_cv.data import (
    CV_DATAMODULE_METHOD_NAME,
    BaseCVDataModule,
    is_cross_val_datamodule,
)
from lightning_cv.fabric import Fabric
from lightning_cv.typehints import (
    KwargType,
    MetricType,
    Number,
    OptimizerConfigT,
    SchedulerConfigT,
)
from lightning_cv.utils.checkpointing import CheckpointingMixin
from lightning_cv.utils.foldutils import (
    convert_output_to_dict,
    flatten_train_val_metrics,
    fold_idiv,
    validation_update,
)
from lightning_cv.utils.stopping import StopReasons
from lightning_cv.utils.strategies import is_distributed

logger = logging.getLogger(__name__)


class TrainerState(enum.Enum):
    TRAINING = enum.auto()
    VALIDATING = enum.auto()
    NOT_STOPPED = enum.auto()


class CrossValidationTrainer(CheckpointingMixin):
    __fabric_keys__ = {
        "accelerator",
        "strategy",
        "devices",
        "precision",
        "plugins",
        "callbacks",
        "loggers",
    }

    __reset_fn__ = partial(torch.fill_, value=0.0)
    __callbacks__ = Callback.available_callbacks()

    def __init__(
        self, model_type: type[LightningModule], config: CrossValidationTrainerConfig
    ):
        self.config = config
        self.model_type = model_type

        self.global_step_per_fold: dict[int, int] = dict()
        """Keeps track of the number of times the model in each fold has been stepped. In other
        words, the number of times the model has been updated by an optimizer."""

        self.current_epoch: int = 0
        self.current_fold: int = 0
        self.setup_complete = False
        self.is_first_epoch = True

        self.should_stop = False
        self.current_train_metrics: MetricType = dict()
        self._current_val_metrics_per_fold: MetricType = dict()

        if self.config.loggers is None:
            self.config.loggers = CSVLogger(
                root_dir=self.config.checkpoint_dir,
            )

        self._add_default_callbacks(self.config.callbacks)

        fabric_kwargs = self.config.to_dict(include=self.__fabric_keys__)
        self.fabric = Fabric(**fabric_kwargs)

        if self.config.gradient_clip_algorithm == "norm":
            self.gradient_clip_kwargs = {"max_norm": self.config.gradient_clip_val}
        else:
            self.gradient_clip_kwargs = {"clip_val": self.config.gradient_clip_val}

        self.apply_gradient_clipping = self.config.gradient_clip_val > 0.0

        self._current_state = TrainerState.TRAINING
        self._state_when_stopped = TrainerState.NOT_STOPPED
        self._status = StopReasons.NULL

    @property
    def num_folds(self) -> int:
        return len(getattr(self, "fold_manager", {}))

    @property
    def current_val_metrics(self) -> MetricType:
        if not hasattr(self, "_current_val_metrics"):
            raise RuntimeError(
                "Current validation metrics aren't available until after the 1st epoch"
            )

        return self._current_val_metrics

    @current_val_metrics.setter
    def current_val_metrics(self, value: MetricType):
        self._current_val_metrics = value

    # TODO: need to distinguish this from the root dir
    # lightning uses: rootdir for lightning_root, and then log_dir for the current version
    # especially since the config uses checkpoint dir
    # and then this is different that
    @cached_property
    def checkpoint_dir(self) -> Path:
        """Get the checkpoint directory for the CURRENT version of training.

        Ex: lighting_root/lightning_logs/version_0
        """
        try:
            logger_expt_name = self.fabric.logger.name or "expt_0"
            logger_version_no = self.fabric.logger.version or "version_0"
            if isinstance(logger_version_no, int):
                logger_version_no = f"version_{logger_version_no}"
            if self.fabric.logger.root_dir is not None:
                checkpoint_dir = Path(self.fabric.logger.root_dir)
            else:
                checkpoint_dir = Path("checkpoints")
        except (AttributeError, IndexError):
            # no logger attached
            logger_expt_name = "expt_0"
            logger_version_no = "version_0"
            checkpoint_dir = self.config.checkpoint_dir

        return checkpoint_dir / logger_expt_name / logger_version_no

    @property
    def is_distributed(self) -> bool:
        return is_distributed(self.fabric.strategy)

    @property
    def status(self) -> StopReasons:
        return self._status

    @status.setter
    def status(self, value: Any):
        if isinstance(value, StopReasons):
            self._status = value
            if self._status != StopReasons.NULL:
                self.state_when_stopped = self.current_state
        else:
            raise TypeError(f"{value=} is not a valid StopReasons enum")

    @property
    def current_state(self) -> TrainerState:
        return self._current_state

    @current_state.setter
    def current_state(self, value: Any):
        if isinstance(value, TrainerState):
            if value == TrainerState.NOT_STOPPED:
                raise ValueError("Current state cannot be NOT_STOPPED")
            self._current_state = value
        else:
            raise TypeError(f"{value=} is not a valid TrainerState enum")

    @property
    def state_when_stopped(self) -> TrainerState:
        return self._state_when_stopped

    @state_when_stopped.setter
    def state_when_stopped(self, value: Any):
        if self.status == StopReasons.NULL or not self.should_stop:
            raise ValueError(
                "Cannot set state_when_stopped when trainer has not been signaled to stop."
            )

        if isinstance(value, TrainerState):
            self._state_when_stopped = value
        else:
            raise TypeError(f"{value=} is not a valid TrainerState enum")

    @property
    def fold_manager(self) -> dict[int, FoldState]:
        if not hasattr(self, "_fold_manager"):
            raise AttributeError(
                "`fold_manager` is not available. The `.setup()` must be called first with a "
                "valid cross validation datamodule and a model config to initialize 1 model per "
                "fold."
            )

        return self._fold_manager

    def _add_default_callbacks(self, callbacks: CallbackT = None):
        callbacks = standardize_callbacks(callbacks)
        callbacks.extend([TrainMode(), EvalMode(), LogFlusher(), TqdmProgressBar()])

        has_model_summary = False
        has_model_checkpoint = False
        for callback in callbacks:
            if isinstance(callback, ModelCheckpoint):
                has_model_checkpoint = True
            elif isinstance(callback, ModelSummary):
                has_model_summary = True

        if not has_model_checkpoint:
            callbacks.append(ModelCheckpoint(save_last=True))

        if not has_model_summary:
            callbacks.append(ModelSummary())

        self.config.callbacks = callbacks

    @property
    def callbacks(self) -> list[Callback]:
        return self.config.callbacks  # type: ignore[return-type] <- should be standardized

    def setup(self, datamodule: BaseCVDataModule, model_config):
        # only setup once
        if not self.setup_complete:
            self.fabric.launch()
            datamodule.setup("fit")

            # setup n_folds models and optimizers
            self._fold_manager: dict[int, FoldState] = dict()
            self.global_step_per_fold.clear()
            self.estimated_steps: dict[int, int] = dict()

            for fold, (train_loader, val_loader) in enumerate(
                datamodule.train_val_dataloaders()
            ):
                self.estimated_steps[fold] = self._estimated_steps(len(train_loader))

                model = self.model_type(model_config)
                setattr(model, "estimated_steps", self.estimated_steps[fold])
                config = self._parse_optimizers_schedulers(model.configure_optimizers())
                optimizer: Optimizer
                optimizer, scheduler_cfg = config

                # ignore fact that fabric returns a wrapper
                # ignore setup takes a torch.nn.Module or a lightning.LightningModule
                # BUG: for some reason fabric will add lightningmodules to the callbacks
                # this was not intended and was added with commit bf2af0e of lightning
                # however, this package was not developed with this in mind
                model, optimizer = self.fabric.setup(model, optimizer)  # type: ignore[return-value]

                self.fold_manager[fold] = FoldState(
                    model=model, optimizer=optimizer, scheduler=scheduler_cfg
                )
                # init global steps for each fold to 0
                self.global_step_per_fold[fold] = 0

            self._current_val_metrics_per_fold["loss"] = torch.zeros(
                size=(len(self.fold_manager),), device=self.fabric.device
            )
            self._finished_cv_loop_per_fold = torch.zeros(
                size=(len(self.fold_manager),), device=self.fabric.device
            ).bool()

            # need to remove model from callbacks since this gets added in later version of lightning
            # this breaks everything...
            callbacks = [
                callback
                for callback in self.fabric._callbacks
                if not isinstance(callback, self.model_type)
            ]

            self.fabric._callbacks = callbacks

            self.setup_complete = True

    def _estimated_steps(self, training_batches: int):
        max_estimated_steps = (
            math.ceil(training_batches / self.config.grad_accum_steps)
            * self.config.max_epochs
        )

        return max_estimated_steps

    def train_with_cross_validation(self, datamodule: BaseCVDataModule, model_config):
        if not is_cross_val_datamodule(datamodule):
            is_lightning_module = isinstance(datamodule, LightningDataModule)
            has_method = hasattr(datamodule, CV_DATAMODULE_METHOD_NAME)

            raise ValueError(
                "The provided datamodule MUST satisfy 2 constraints:\n"
                "\t1. It must be a subclass of lightning.LightningDataModule\n"
                "\t2. It must have a method train_val_dataloaders() that yields pairs of "
                "dataloaders for training and validation."
                f"Provided module is a LightningDataModule: {is_lightning_module}\n"
                f"Provided module has method {CV_DATAMODULE_METHOD_NAME}: {has_method}"
            )

        # set up n_folds models and optimizers, and creates fold_manager attribute
        # also calls self.fabric.launch()
        self.setup(datamodule=datamodule, model_config=model_config)
        self.apply_callback("on_train_start")
        while not self.should_stop:
            # calls train_loop, val_loop, step_scheduler per fold
            self.cross_val_loop(datamodule=datamodule)

            self.current_epoch += 1

            self.is_first_epoch = False

            # stopping condition on epoch level
            # don't change self.should_stop here
            # reserve self.should_stop for external signals for stopping
            if self.current_epoch >= self.config.max_epochs:
                break

        self.apply_callback("on_train_end")

        # reset for next fit call
        self.should_stop = False

        self.log_status()

    def cross_val_loop(self, datamodule: BaseCVDataModule):
        # called once per epoch
        # train and validate each fold
        self.apply_callback("on_train_fold_start")

        # reset here - advt is once training is complete, this will keep history of
        # per fold completeness
        self._finished_cv_loop_per_fold.fill_(False)

        foldloaders = datamodule.train_val_dataloaders(shuffle=True)
        for fold, (train_loader, val_loader) in enumerate(foldloaders):
            if self.should_stop:
                break

            train_loader, val_loader = self.fabric.setup_dataloaders(
                train_loader, val_loader
            )

            self.current_fold = fold
            state = self.fold_manager[fold]
            # ignore fact that lightning.fabric wraps model
            model = cast(LightningModule, state.model)
            optimizer = state.optimizer
            scheduler_cfg = state.scheduler

            self.train_loop(
                model=model,
                optimizer=optimizer,
                train_loader=train_loader,
                scheduler_cfg=scheduler_cfg,
            )

            self.val_loop(model=model, val_loader=val_loader, optimizer=optimizer)

            self.step_scheduler(
                model=model,
                scheduler_cfg=scheduler_cfg,
                level="epoch",
                current_value=self.current_epoch,
            )

            self._finished_cv_loop_per_fold[fold] = True
        # update this value every epoch
        # this stores the average validation loss per fold

        if self.is_distributed:  # pragma: no cover <- TODO: need to test distributed
            # need to get all metrics in distributed setting
            val_metrics = cast(
                MetricType, self.fabric.all_gather(self._current_val_metrics_per_fold)
            )
        else:
            val_metrics = self._current_val_metrics_per_fold

        # BUG: for some reason this doesn't work
        # in CHTC trials stopped by timeouts are averaging as if
        # all were completed...
        # however, this isn't the case with a local test
        # so likely installation issue...
        # most likely timer issue, but early stopping log messages may help diagnose
        if self._finished_cv_loop_per_fold.all():
            # only update if all folds finished training and validating
            # this will technically raise on error on the first epoch since
            # self.current_val_metrics is not yet available -> but that's already
            # indicates some problem anyway. Otherwise, the current_val_metrics
            # will just be the previous val metrics. This will only happen
            # when an external callback stops training early.
            #
            # The only callback that has the ability to stop training other than
            # at the end of this fn with the "on_train_fold_end" callback is the
            # Timer callback, and not entering this block basically indicates
            # training is about to stop anyway.

            # current_val_metrics = average over folds
            self.current_val_metrics = apply_to_collection(
                val_metrics, dtype=torch.Tensor, function=torch.mean
            )

        # reset values
        apply_to_collection(
            self._current_val_metrics_per_fold,
            dtype=torch.Tensor,
            function=self.__reset_fn__,
        )

        self.apply_callback("on_train_fold_end")

    def train_loop(
        self,
        model: LightningModule,
        optimizer: Optimizer,
        train_loader: DataLoader,
        scheduler_cfg: Optional[SchedulerConfigT] = None,
    ):
        self.current_state = TrainerState.TRAINING
        n_batches = len(train_loader)
        batch_limit = self._limit_batches(n_batches, self.config.limit_train_batches)

        self.apply_callback(
            "on_train_epoch_start_per_fold",
            model=model,
            total=batch_limit,
            current_epoch=self.current_epoch,
            current_fold=self.current_fold,
        )  # calls model.train()

        # default starting value in case loop never runs
        output = flatten_train_val_metrics(
            train_metrics=self.current_train_metrics,
            val_metrics=self.current_val_metrics if not self.is_first_epoch else None,
            prepend_stage=not self.is_first_epoch,
        )
        for batch_idx, batch in enumerate(train_loader):
            # end epoch if stopping training completely or
            # max batches for this epoch reached
            if self.should_stop or batch_idx >= batch_limit:
                break

            self.apply_callback(
                "on_train_batch_start_per_fold", batch=batch, batch_idx=batch_idx
            )

            # check if optimizer should step in gradient accumulation
            should_optim_step = self.should_do_something(
                batch_idx, self.config.grad_accum_steps
            )

            if should_optim_step:
                # currently only supports a single optimizer
                self.apply_callback(
                    "on_before_optimizer_step", optimizer=optimizer, optimizer_idx=0
                )

                # although some optimizers need a closure
                # these are not compatibile with automatic precision scaling
                # .training_step sets `self.current_train_metrics`
                self.training_step(model=model, batch=batch, batch_idx=batch_idx)
                if self.apply_gradient_clipping:  # pragma: no cover <- this is tested
                    self.fabric.clip_gradients(
                        module=model,
                        optimizer=optimizer,
                        error_if_nonfinite=False,
                        **self.gradient_clip_kwargs,
                    )

                # optimizer.step() # TODO: model.optimizer_step() instead?
                model.optimizer_step(
                    epoch=self.current_epoch, batch_idx=batch_idx, optimizer=optimizer
                )

                self.apply_callback("on_before_zero_grad", optimizer=optimizer)

                # optimizer.zero_grad()
                model.optimizer_zero_grad(
                    epoch=self.current_epoch, batch_idx=batch_idx, optimizer=optimizer
                )

                metrics = {
                    "epoch": self.current_epoch,
                    "fold": self.current_fold,
                    "stage": "train",
                    **self.current_train_metrics,
                }

                self.apply_callback(
                    "on_before_log_metrics", metrics=metrics, optimizer=optimizer
                )
                self.fabric.log_dict(
                    metrics=metrics,
                    step=self.global_step_per_fold[self.current_fold],
                )

                self.step_scheduler(
                    model,
                    scheduler_cfg,
                    level="step",
                    current_value=self.global_step_per_fold[self.current_fold],
                )

            else:
                # gradient accumulation -> no optimizer step
                self.training_step(model=model, batch=batch, batch_idx=batch_idx)

            # after first epoch, output now has both the current training
            # AND validation metrics
            val_metrics = self.current_val_metrics if not self.is_first_epoch else None
            output = flatten_train_val_metrics(
                train_metrics=self.current_train_metrics,
                val_metrics=val_metrics,
                prepend_stage=not self.is_first_epoch,
            )

            self.apply_callback(
                "on_train_batch_end_per_fold",
                output=output,
                batch=batch,
                batch_idx=batch_idx,
            )

            # only increase global step if optimizer stepped
            self.global_step_per_fold[self.current_fold] += int(should_optim_step)

        self.apply_callback("on_train_epoch_end_per_fold", output=output)

    def training_step(
        self, model: LightningModule, batch: Any, batch_idx: int
    ) -> torch.Tensor:
        _outputs = model.training_step(batch, batch_idx=batch_idx)
        if _outputs is None:
            raise RuntimeError(
                "It is currently expected that the model's training_step returns the loss and "
                "optionally other metrics. Manual optimization is not currently supported."
            )

        outputs: MetricType = convert_output_to_dict(_outputs)

        loss = outputs["loss"]

        self.apply_callback("on_before_backward", loss=loss)
        self.fabric.backward(loss)
        self.apply_callback("on_after_backward")

        # avoid gradients in stored/accumulated values -> prevents potential OOM
        self.current_train_metrics = cast(
            MetricType,
            apply_to_collection(outputs, dtype=torch.Tensor, function=torch.detach_),
        )

        return loss

    def step_scheduler(
        self,
        model: LightningModule,
        scheduler_cfg: Optional[SchedulerConfigT],
        level: Literal["step", "epoch"],
        current_value: int,
    ) -> None:
        # no scheduler
        if scheduler_cfg is None:
            return

        # wrong interval (step vs. epoch)
        if scheduler_cfg["interval"] != level:
            return

        # right interval, but wrong step wrt frequency
        if current_value % cast(int, scheduler_cfg["frequency"]) != 0:
            return

        # if self.current_epoch == 0:
        #     return

        # ONLY allow schedulers to step per epoch and keep all fold schedulers synced
        # TODO: but this code is called from train_loop, which gives the option to step per batch
        # not sure if it's best to allow LR to change each fold?
        # I guess it's fine since each fold would always have its LR
        # adjusted the same across experiments

        if self.current_epoch == 0:
            current_val_metrics = None
        else:
            current_val_metrics = self.current_val_metrics

        possible_monitor_vals = flatten_train_val_metrics(
            train_metrics=self.current_train_metrics,
            val_metrics=current_val_metrics,
            prepend_stage=True,
        )

        try:
            monitor = possible_monitor_vals[cast(str, scheduler_cfg["monitor"])]
        except KeyError as ex:
            possible_keys = list(possible_monitor_vals.keys())
            raise KeyError(
                f"monitor {scheduler_cfg['monitor']} is invalid. Possible values "
                f"are {possible_keys}."
            ) from ex

        # rely on model hook for actual step
        scheduler = cast(LRScheduler, scheduler_cfg["scheduler"])
        model.lr_scheduler_step(scheduler, monitor)

    @staticmethod
    def _limit_batches(n_batches: int, limit_batches: Number = 1.0) -> int:
        if not isinstance(limit_batches, int):
            # is a float
            batch_limit = round(n_batches * limit_batches)
        else:
            batch_limit = limit_batches

        return batch_limit

    @property
    def should_validate(self) -> bool:
        """Whether to currently run validation."""
        return self.should_do_something(
            self.current_epoch, self.config.validation_frequency
        )

    def should_do_something(self, current_step_0_indexed: int, frequency: int) -> bool:
        # if we want to do something with a frequency of 2 steps, then we need to check
        # the +1 of the current 0-indexed step:
        # step: 0 1 2 3 4 ...
        # do  : F T F T F ...
        # this prevents the first step (0) from always being True
        return (current_step_0_indexed + 1) % frequency == 0

    @torch.no_grad()
    def val_loop(
        self,
        model: LightningModule,
        val_loader: DataLoader,
        optimizer: Optimizer,
    ):
        """The validation loop running a single validation epoch.

        Args:
            model: the LightningModule to evaluate
            val_loader: The dataloader yielding the validation batches.
        """
        self.current_state = TrainerState.VALIDATING

        n_batches = len(val_loader)
        batch_limit = self._limit_batches(n_batches, self.config.limit_val_batches)

        self.apply_callback(
            "on_validation_start_per_fold", model=model, total=batch_limit
        )  # calls `model.eval()`

        batch_idx = 0  # for mypy unbound local error
        for batch_idx, batch in enumerate(val_loader):
            # end epoch if stopping training completely or
            # max batches for this epoch reached
            if self.should_stop or batch_idx >= batch_limit:
                break

            self.apply_callback(
                "on_validation_batch_start_per_fold",
                batch=batch,
                batch_idx=batch_idx,
                current_fold=self.current_fold,
            )

            _out = model.validation_step(batch, batch_idx=batch_idx)
            if _out is None:
                raise RuntimeError(
                    "The model's validation_step is expected to return the loss and "
                    "optionally other metrics."
                )

            out: MetricType = convert_output_to_dict(_out)

            metrics = {
                "epoch": self.current_epoch,
                "fold": self.current_fold,
                "stage": "val",
                **out,
            }

            self.apply_callback(
                "on_before_log_metrics", metrics=metrics, optimizer=optimizer
            )
            self.fabric.log_dict(
                metrics=metrics,
                step=self.global_step_per_fold[self.current_fold] - 1,
            )

            self.apply_callback(
                "on_validation_batch_end_per_fold",
                output=out,
                batch=batch,
                batch_idx=batch_idx,
            )

            # takes all metrics reported in out and add in-place
            # to the tmp storage at current_val_metrics_per_fold
            # then later divide by number of steps taken to avg
            validation_update(
                current_metrics=self._current_val_metrics_per_fold,
                output=out,
                fold=self.current_fold,
            )

        # avg fold level loss by number of batches/steps taken
        apply_to_collection(
            data=self._current_val_metrics_per_fold,
            dtype=torch.Tensor,
            function=partial(
                fold_idiv,
                # in most cases the denom will just be the number of batches
                # however, if using limit_val_batches, it will be the number of batches
                # actually stepped, but occasionally if an external callback like
                # a timer stops training, then the number of batches stepped will be
                # less than the batch_limit. The batch_idx will then be the number
                # of batches stepped. We need to +1 to avoid div by 0
                # if in this scenario and no batches were stepped.
                value=min(batch_limit, batch_idx + 1),
                fold=self.current_fold,
            ),
        )

        self.apply_callback("on_validation_end_per_fold")

    def _parse_optimizers_schedulers(
        self,
        configure_optim_output: Union[OptimizerLRScheduler, Mapping],
    ) -> OptimizerConfigT:
        # NOTE: none of this supports multiples optimizers or schedulers

        # single optimizer
        if isinstance(configure_optim_output, Optimizer):
            return configure_optim_output, None

        _lr_sched_defaults = {
            "interval": "epoch",
            "frequency": 1,
            "monitor": self.config.monitor,
        }

        # single lr scheduler config with optimizer passed
        if isinstance(configure_optim_output, Mapping):
            optimizer = configure_optim_output["optimizer"]
            lr_scheduler: KwargType = configure_optim_output.get("lr_scheduler", None)  # type: ignore
            if lr_scheduler is not None:
                lr_scheduler = _lr_sched_defaults | lr_scheduler
            return optimizer, lr_scheduler

        # tuple of optimizer and lr scheduler
        # this would technically handle multiple optimizers and schedulers
        if isinstance(configure_optim_output, tuple):
            optimizer, lr_scheduler = configure_optim_output  # type: ignore

            if isinstance(lr_scheduler, Mapping):
                lr_scheduler = _lr_sched_defaults | lr_scheduler
            else:
                lr_scheduler = {"scheduler": lr_scheduler, **_lr_sched_defaults}
            return optimizer, lr_scheduler  # type: ignore[return-value]

        raise ValueError(
            "Output of model.configure_optimizers() must be a torch.optim.Optimizer "
            "or a Mapping with keys 'optimizer' and optionally 'lr_scheduler'"
        )

    @property
    def is_global_zero(self) -> bool:
        return self.fabric.is_global_zero

    def apply_callback(self, callback_name: str, **kwargs):
        if callback_name not in self.__callbacks__:
            raise RuntimeError(
                f"Callback {callback_name} not found. "
                "Allowed callbacks are: {self.__callbacks__}"
            )
        # all callbacks have a trainer kwarg
        kwargs["trainer"] = self
        self.fabric.call(callback_name, **kwargs)

    def log_status(self):
        if self.is_global_zero:
            msg_template = self.status.message()
            msg_kwargs = {
                "max_epochs": self.config.max_epochs,
                "monitor": self.config.monitor,
            }
            # str.format only takes what is needed
            msg = msg_template.format(**msg_kwargs)
            logger.info(f"{msg}\n")



from copy import deepcopy
from pathlib import Path
from typing import Callable, Mapping, Optional, Union, cast

import optuna
from lightning import LightningModule
from lightning.fabric.loggers.csv_logs import CSVLogger
from lightning.fabric.loggers.logger import Logger
from optuna.trial import TrialState

from lightning_cv.callbacks.utils import standardize_callbacks
from lightning_cv.config import CrossValidationTrainerConfig
from lightning_cv.data import BaseCVDataModule
from lightning_cv.trainer import CrossValidationTrainer
from lightning_cv.tuning.callbacks import TrialPruning
from lightning_cv.tuning.hyperparameters import HparamRegistry, IntFloatStrDict, suggest
from lightning_cv.tuning.utils import _TrialUpdater
from lightning_cv.typehints import FilePath
from lightning_cv.typehints.config import ConfigT
from lightning_cv.utils.serde import config_deserializer, config_serializer

TrialSuggestionFn = Callable[[optuna.Trial], IntFloatStrDict]


class Tuner:
    current_trainer: CrossValidationTrainer

    def __init__(
        self,
        model_type: type[LightningModule],
        model_config: ConfigT,
        datamodule_type: type[BaseCVDataModule],
        datamodule_config: ConfigT,
        trainer_config: CrossValidationTrainerConfig,
        logdir: FilePath = Path("checkpoints"),
        experiment_name: str = "exp0",
        clean_logdir_if_exists: bool = False,
        verbose: bool = False,
        hparam_config_file: Optional[FilePath] = None,
    ):
        self._hparam_registry = self._init_hparam_registry(hparam_config_file)

        self.model_type = model_type
        # only store dict and type so that it can be updated with new trialed values
        serialized_model_config = config_serializer(model_config)
        self.model_config_type = type(model_config)

        self.datamodule_type = datamodule_type
        self.data_config_type = type(datamodule_config)
        serialized_data_config = config_serializer(datamodule_config)

        self.trial_updater = _TrialUpdater(
            model=serialized_model_config, data=serialized_data_config
        )

        self.verbose = verbose
        self.experiment_name = experiment_name
        self.trainer_config = trainer_config.copy(deep=True)
        self._init_logdir(logdir, clean_logdir_if_exists)

        self._callbacks = standardize_callbacks(self.trainer_config.callbacks)
        self._logger_types = self._store_logger_types(self.trainer_config.loggers)

        self._trial_number = 0

    def _init_hparam_registry(
        self, config_file: Optional[FilePath]
    ) -> Optional[HparamRegistry]:
        if config_file is None:
            return None

        return HparamRegistry.from_file(config_file)

    def _init_logdir(self, logdir: FilePath, clean_logdir_if_exists: bool = False):
        if isinstance(logdir, str):
            logdir = Path(logdir)

        self.logdir = logdir

        # the lightning loggers log to:
        # (default_root_dir/name/version_#)
        # so need to check if default_root_dir/name exists
        actual_path = self.logdir.joinpath(self.experiment_name)
        # if actual_path.exists():
        #     if not clean_logdir_if_exists:
        #         raise RuntimeError(
        #             f"The log directory {self.logdir} exists. Use "
        #             "`clean_logdir_if_exists` or change the log directory."
        #         )
        #     elif self.trainer_config.devices > 1:
        #         rmtree(actual_path)

        actual_path.mkdir(exist_ok=True, parents=True)

    def _store_logger_types(
        self, loggers: Union[list[Logger], Logger, None]
    ) -> list[type[Logger]]:
        # capture logger types only and reconstruct during tuning
        if loggers is not None:
            if not isinstance(loggers, list):
                logger_types = [type(loggers)]
            else:
                logger_types = [type(logger) for logger in loggers]
        else:
            logger_types = cast(list[type[Logger]], [CSVLogger])

        return logger_types

    def _create_datamodule(self) -> BaseCVDataModule:
        # these should have already been updated, but check just in case
        self._check_hparams_updated()
        serialized_data_config = self.trial_updater.data
        data_config = config_deserializer(serialized_data_config, self.data_config_type)

        if isinstance(data_config, Mapping):
            try:
                datamodule = self.datamodule_type(**data_config)
            except TypeError:
                # datamodule might expect the config as a single argument
                # which is what is tried next
                pass
            else:
                return datamodule

        # otherwise datamodule must accept only a single config argument, even if it could
        # be a dict
        return self.datamodule_type(data_config)

    def _create_model_config(self) -> ConfigT:
        # these should have already been updated, but check just in case
        self._check_hparams_updated()
        serialized_model_config = self.trial_updater.model
        model_config = config_deserializer(
            serialized_model_config, self.model_config_type
        )

        return model_config

    def _check_hparams_updated(self):
        if not self.trial_updater._updated:
            raise RuntimeError(
                "The tunable hyperparameters have not yet been updated with new trialed "
                "values. Try calling the `.trial_updater.update` method."
            )

    def _set_trial_number(self, study: optuna.Study):
        trials = study.get_trials(deepcopy=False)
        if trials:
            latest_trial = max(trials, key=lambda x: x.number)
            if latest_trial.state == TrialState.RUNNING:
                # latest_trial is actively running so use that number
                number = latest_trial.number
            else:
                # latest_trial completed so need to increment
                number = latest_trial.number + 1
        else:
            # default currently stored value which starts at 0
            # should only be used for the first trial
            # then the number should always be handled by the optuna.Study
            number = self._trial_number
        return number

    @property
    def trial_number(self) -> int:
        return self._trial_number

    # TODO: this is very heavily coupled with optuna... may want to try a refactor to add support
    # for a few diff libs? particularly ray.tune since that supports multi GPU tuning
    def tune(
        self,
        trial: optuna.Trial,
        suggest_fn: Optional[TrialSuggestionFn] = None,
        trialed_values: Optional[IntFloatStrDict] = None,
        monitor: str = "loss",
    ) -> float:
        self._trial_number = self._set_trial_number(trial.study)
        if trialed_values is None:
            # if not trialed_values passed, then need to check for a suggestion fn
            if suggest_fn is None:
                if self._hparam_registry is None:
                    raise ValueError(
                        "Either (1) a hparam config toml file must be passed upon Tuner "
                        "init, (2) a suggestion function must be passed to this method, "
                        "or (3) a custom trialed_value dictionary must be passed."
                    )

                # use HparamRegistry functionality
                trialed_values = suggest(trial=trial, registry=self._hparam_registry)
            else:
                # use the provided suggestion fn
                trialed_values = suggest_fn(trial)
        # else: trialed_values must've been passed

        loggers = [
            logger_type(
                root_dir=self.logdir,  # type: ignore
                name=self.experiment_name,  # type: ignore
                version=self._trial_number,  # type: ignore
            )
            for logger_type in self._logger_types
        ]

        self.trainer_config.loggers = loggers

        callbacks = deepcopy(self._callbacks)
        callbacks.append(
            TrialPruning(trial=trial, monitor=monitor, verbose=self.verbose)
        )

        self.trainer_config.callbacks = callbacks

        self.current_trainer = CrossValidationTrainer(
            model_type=self.model_type, config=self.trainer_config
        )

        # update model/data config with trialed values
        self.trial_updater.update(trialed_values)

        # then update model config
        model_config = self._create_model_config()
        datamodule = self._create_datamodule()

        try:
            self.current_trainer.train_with_cross_validation(
                datamodule=datamodule, model_config=model_config
            )
        finally:
            # always increment trial number even if trial gets pruned
            # NOTE: this shouldn't be needed since this will mostly defer
            # to the `optuna.Study` update its trial numbers
            self._trial_number += 1

            self.current_trainer.log_status()
            self.trainer_config.callbacks = None

        return self.current_trainer.current_val_metrics[monitor].item()



import time
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Literal, Union

from lightning_cv.callbacks.base import Callback
from lightning_cv.callbacks.utils import remove_old_checkpoints
from lightning_cv.typehints import MetricType
from lightning_cv.utils.stopping import StopReasons

if TYPE_CHECKING:
    from lightning_cv.trainer import CrossValidationTrainer


Intervals = Literal["epoch", "step", "fold", "immediately"]


class Timer(Callback):
    TIMEORDER = ("days", "hours", "minutes", "seconds")

    def __init__(
        self,
        duration: Union[str, timedelta, dict[str, int], None] = None,
        buffer: Union[timedelta, dict[str, int], None] = None,
        interval: Intervals = "immediately",
        save_latest: bool = True,
        training_only: bool = False,
        verbose: bool = True,
    ) -> None:
        if isinstance(duration, str):
            duration = Timer._str_to_timedelta_dict(duration)

        if isinstance(duration, dict):
            duration = Timer._dict_to_timedelta(duration)

        if isinstance(buffer, dict):
            buffer = Timer._dict_to_timedelta(buffer)

        self._start_time = time.monotonic()
        self._interval = interval
        self._verbose = verbose
        self._save_latest = save_latest
        cls_name = self.__class__.__name__
        self._repr_fields: list[str] = list()
        if duration is not None:
            self._repr_fields.append(f"duration={repr(duration)}")
            if buffer is not None:
                duration -= buffer
                self._repr_fields.append(f"buffer={repr(buffer)}")
            self._duration = duration.total_seconds()
        else:
            self._duration = None

        self._repr = f"{cls_name}({', '.join(self._repr_fields)})"

        self._check_during_validation = not training_only
        self._current_fold_training_complete = False
        self._stopped = False

    @staticmethod
    def _str_to_timedelta_dict(timestr: str) -> dict[str, int]:
        n_time_units = len(Timer.TIMEORDER)

        times = timestr.strip().split(":")
        n_times = len(times)
        if n_times < 2 or n_times > n_time_units:
            raise ValueError(
                f"Invalid time string: {timestr}. "
                "Should pass DD:HH:MM:SS, HH:MM:SS, or MM:SS"
            )

        start_idx = n_time_units - n_times
        units = Timer.TIMEORDER[start_idx:]

        return {unit: int(intv) for intv, unit in zip(times, units)}

    @staticmethod
    def _dict_to_timedelta(time: dict[str, int]) -> timedelta:
        return timedelta(**time)

    def _elapsed_time(self) -> float:
        return time.monotonic() - self._start_time

    def _is_time_up(self) -> bool:
        if self._duration is None:
            return False
        return self._elapsed_time() >= self._duration

    def _check_time(self, trainer: "CrossValidationTrainer"):
        should_stop = self._is_time_up()
        self._stopped = should_stop

        if trainer.is_distributed:  # pragma: no cover <- can't test distributed yet
            should_stop = trainer.fabric.strategy.reduce_boolean_decision(
                should_stop, all=False
            )

        trainer.should_stop = trainer.should_stop or should_stop
        if should_stop:  # pragma: no cover <- this is actually tested
            trainer.status = StopReasons.TIMEOUT

    def _check_time_and_checkpoint(self, trainer: "CrossValidationTrainer"):
        self._check_time(trainer)
        if trainer.should_stop and self._save_latest:
            remove_old_checkpoints(trainer.checkpoint_dir, "*latest*.ckpt")
            save_kwargs: dict[str, Any] = dict(suffix="latest", include_val_loss=False)
            for fold, state in trainer.fold_manager.items():
                save_kwargs["states"] = state
                save_kwargs["fold"] = fold

                # simplified since trainer now keeps track of this
                save_kwargs["training_complete"] = bool(
                    trainer._finished_cv_loop_per_fold[fold]
                )

                trainer.save(**save_kwargs)

    def _validate_interval_and_check(
        self,
        trainer: "CrossValidationTrainer",
        interval: Union[set[Intervals], Intervals],
    ):
        if isinstance(interval, str):  # pragma: no cover
            interval = {interval}

        if self._interval not in interval:
            return

        if not self._stopped:
            # need to guard here with saving bc if stopping at batch level
            # could start upper level loops and save again
            self._check_time_and_checkpoint(trainer)

    def on_train_fold_start(self, trainer: "CrossValidationTrainer"):
        self._current_fold_training_complete = False

    def on_train_fold_end(self, trainer: "CrossValidationTrainer"):
        self._validate_interval_and_check(trainer, {"fold", "immediately"})

    def on_train_epoch_end_per_fold(
        self, trainer: "CrossValidationTrainer", output: MetricType
    ):
        self._current_fold_training_complete = True
        self._validate_interval_and_check(trainer, {"epoch", "immediately"})

    def on_train_batch_end_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        output: MetricType,
        batch,
        batch_idx: int,
    ):
        self._validate_interval_and_check(trainer, {"step", "immediately"})

    def on_validation_batch_end_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        output: MetricType,
        batch,
        batch_idx: int,
    ):
        if self._check_during_validation:
            self._validate_interval_and_check(trainer, {"step", "immediately"})

    def on_validation_end_per_fold(self, trainer: "CrossValidationTrainer"):
        if self._check_during_validation:
            self._validate_interval_and_check(trainer, {"epoch", "immediately"})

    def __repr__(self) -> str:
        return self._repr



from typing import TYPE_CHECKING, TypeVar, Union

from torch import Tensor
from torch.optim import Optimizer

from lightning_cv.typehints import KwargType, MetricType

if TYPE_CHECKING:
    from lightning import Callback as LightingCallback
    from lightning import LightningModule

    from lightning_cv.trainer import CrossValidationTrainer


class Callback:
    """Simple interface for supported `lightning.Fabric` callbacks during
    cross validation."""

    def on_train_start(self, trainer: "CrossValidationTrainer"): ...

    def on_train_end(self, trainer: "CrossValidationTrainer"): ...

    def on_train_fold_start(self, trainer: "CrossValidationTrainer"): ...

    def on_train_fold_end(self, trainer: "CrossValidationTrainer"): ...

    def on_train_epoch_start_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        model: "LightningModule",
        total: int,
        current_epoch: int,
        current_fold: int,
    ): ...

    def on_train_epoch_end_per_fold(
        self, trainer: "CrossValidationTrainer", output: MetricType
    ): ...

    def on_train_batch_start_per_fold(
        self, trainer: "CrossValidationTrainer", batch, batch_idx: int
    ): ...

    def on_train_batch_end_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        output: MetricType,
        batch,
        batch_idx: int,
    ): ...

    def on_before_optimizer_step(
        self,
        trainer: "CrossValidationTrainer",
        optimizer: Optimizer,
        optimizer_idx: int = 0,
    ): ...

    def on_before_zero_grad(
        self, trainer: "CrossValidationTrainer", optimizer: Optimizer
    ): ...

    def on_validation_start_per_fold(
        self, trainer: "CrossValidationTrainer", model: "LightningModule", total: int
    ): ...

    def on_validation_end_per_fold(self, trainer: "CrossValidationTrainer"): ...

    def on_before_log_metrics(
        self,
        trainer: "CrossValidationTrainer",
        metrics: KwargType,
        optimizer: Optimizer,
    ): ...

    def on_validation_batch_start_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        batch,
        batch_idx: int,
        current_fold: int,
    ): ...

    def on_validation_batch_end_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        output: MetricType,
        batch,
        batch_idx: int,
    ): ...

    def on_before_backward(self, trainer: "CrossValidationTrainer", loss: Tensor): ...

    def on_after_backward(self, trainer: "CrossValidationTrainer"): ...

    def __repr__(self) -> str:  # pragma: no cover

        return f"{self.__class__.__name__}()"

    @classmethod
    def available_callbacks(cls) -> set[str]:
        """Returns the set of available callbacks."""
        return {method for method in dir(cls) if method.startswith("on_")}


_CallbackT = TypeVar("_CallbackT", Callback, "LightningModule", "LightingCallback")
CallbackT = Union[list[_CallbackT], _CallbackT, None]



import logging
from typing import TYPE_CHECKING, Callable, Literal, Optional

import torch

from lightning_cv.callbacks.base import Callback
from lightning_cv.typehints import MetricType
from lightning_cv.utils.stopping import StopReasons

if TYPE_CHECKING:
    from lightning_cv.trainer import CrossValidationTrainer


ComparisonOp = Callable[..., torch.Tensor]
logger = logging.getLogger(__name__)


# mostly copied from lightning.pytorch.callbacks.early_stopping.EarlyStopping
class EarlyStopping(Callback):
    mode_dict: dict[str, ComparisonOp] = {"min": torch.lt, "max": torch.gt}
    order_dict = {"min": "<", "max": ">"}

    def __init__(
        self,
        monitor: str = "loss",
        min_delta: float = 0.0,
        patience: int = 3,
        mode: Literal["min", "max"] = "min",
        check_finite: bool = True,
        stopping_threshold: Optional[float] = None,
        divergence_threshold: Optional[float] = None,
        verbose: bool = True,
        # check on train_epoch end -> think no
    ):
        self.monitor = monitor
        self.min_delta = min_delta
        self.patience = patience
        self.mode = mode
        self.check_finite = check_finite
        self.stopping_threshold = stopping_threshold
        self.divergence_threshold = divergence_threshold
        self.verbose = verbose
        self.wait_count = 0
        self.stopped_epoch = 0

        if mode not in self.mode_dict:
            allowed_values = ", ".join(self.mode_dict.keys())
            raise RuntimeError(f"Passed {mode=}, but must be one of {allowed_values}")

        # TODO: why min_delta thing
        inf_tensor = torch.tensor(torch.inf)
        if self.monitor_op == torch.gt:
            # max monitored value
            self.min_delta *= 1  # needs to increase
            self.best_score = -inf_tensor
        else:
            # min monitored value
            self.min_delta *= -1  # needs to decrease
            self.best_score = inf_tensor

    @property
    def monitor_op(self) -> ComparisonOp:
        return self.mode_dict[self.mode]

    def on_train_fold_end(self, trainer: "CrossValidationTrainer"):
        self._run_early_stopping_check(trainer)

    def _run_early_stopping_check(self, trainer: "CrossValidationTrainer"):
        metrics = trainer.current_val_metrics
        current = self._get_current_metric(metrics)
        should_stop, reason = self._evaluate_stopping_criteria(current)

        if trainer.is_distributed:  # pragma: no cover
            # stop if ANY processes decide to stop
            should_stop = trainer.fabric.strategy.reduce_boolean_decision(
                should_stop, all=False
            )

        trainer.should_stop = trainer.should_stop or should_stop
        if should_stop:
            self.stopped_epoch = trainer.current_epoch
            trainer.status = StopReasons.NO_IMPROVEMENT

        if reason and self.verbose:
            self._log_info(trainer, reason)

    def _get_current_metric(self, metrics: MetricType) -> torch.Tensor:
        try:
            current = metrics[self.monitor]
        except KeyError as err:
            raise RuntimeError(
                f"Monitor value {self.monitor} is not reported in validation metrics: "
                f"{metrics}"
            ) from err
        return current

    def _evaluate_stopping_criteria(
        self, current: torch.Tensor
    ) -> tuple[bool, Optional[str]]:
        should_stop = False
        reason = None
        order_str = self.order_dict[self.mode]

        if self.check_finite and not torch.isfinite(current):
            should_stop = True
            reason = (
                f"Monitored metric {self.monitor} = {current} is not finite. Previous "
                f"best value was {self.best_score:.3f}. Signaling Trainer to stop."
            )
        elif self.stopping_threshold is not None and self.monitor_op(
            current, self.stopping_threshold
        ):
            should_stop = True
            reason = (
                "Stopping threshold reached: "
                f"{self.monitor} = {current} {order_str} {self.stopping_threshold}. "
                "Signaling Trainer to stop."
            )
        elif self.divergence_threshold is not None and self.monitor_op(
            -current, self.divergence_threshold
        ):
            should_stop = True
            reason = (
                "Divergence threshold reached: "
                f"{self.monitor} = {current} {order_str} {self.divergence_threshold}. "
                "Signaling Trainer to stop."
            )
        elif self.monitor_op(
            current - self.min_delta, self.best_score.to(current.device)
        ):
            ### RESET counter back to 0, ie there was improvement
            should_stop = False
            reason = self._improvement_message(current)
            self.best_score = current
            self.wait_count = 0
        else:
            self.wait_count += 1
            if self.wait_count >= self.patience:
                should_stop = True
                reason = (
                    f"Monitored metric {self.monitor} did not improve in the last "
                    f"{self.wait_count} records. Best score: {self.best_score:.3f}. "
                    "Signaling Trainer to stop."
                )

        return should_stop, reason

    def _improvement_message(self, current: torch.Tensor) -> str:
        if torch.isfinite(self.best_score):
            improvement = abs(self.best_score - current)
            msg = (
                f"Metric {self.monitor} improved by {improvement:.3f} >= min_delta "
                f"= {abs(self.min_delta)}. New best score: {current:.3f}"
            )
        else:
            msg = f"Metric {self.monitor} improved. New best score: {current:.3f}"
        return msg

    @staticmethod
    def _log_info(trainer: "CrossValidationTrainer", message: str):
        if trainer.is_global_zero:  # pragma: no cover
            logger.info(f"[rank: 0]: {message}")



import sys
from dataclasses import Field
from typing import ClassVar, Protocol, Union
from attrs import AttrsInstance

from lightning_cv.typehints import KwargType

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field]]


class PydanticBaseModel(Protocol):
    def dict(self) -> KwargType: ...
    def model_dump(self) -> KwargType: ...
    def model_validate(self, value: KwargType) -> Self: ...
    def validate(self, value: KwargType) -> Self: ...


ConfigT = Union[KwargType, AttrsInstance, DataclassInstance, PydanticBaseModel]



from collections.abc import Mapping
from typing import Literal, Optional, Union, cast

from torch import Tensor

from lightning_cv.typehints import MetricType, Number
from lightning_cv.typehints.module import ForwardOutputType

FoldOutputValue = Union[Number, Tensor]


def fold_idiv(per_fold_tensor: Tensor, value: FoldOutputValue, fold: int):
    assert value != 0.0
    per_fold_tensor[fold] /= value


def fold_iadd(per_fold_tensor: Tensor, value: FoldOutputValue, fold: int):
    per_fold_tensor[fold] += value


def output_getter(out: ForwardOutputType, name: str = "loss") -> Tensor:
    if isinstance(out, Tensor):
        return out

    return out[name]


def validation_update(
    current_metrics: MetricType, output: Mapping[str, Tensor], fold: int
):
    for metric, value in output.items():
        if metric not in current_metrics:
            # this will exist bc it is created during trainer setup
            per_fold_loss_tensor = current_metrics["loss"]
            n_folds = int(per_fold_loss_tensor.size(0))

            # created the empty tensor to store values in for each fold
            current_metrics[metric] = per_fold_loss_tensor.new_zeros(size=(n_folds,))

        fold_iadd(per_fold_tensor=current_metrics[metric], value=value, fold=fold)


def convert_output_to_dict(metrics: ForwardOutputType) -> MetricType:
    if isinstance(metrics, Tensor):
        # metric is just loss tensor
        metrics = {"loss": metrics}
    return cast(MetricType, metrics)


def prepend_stage_to_metrics(
    stage: Literal["train", "val"], metrics: MetricType
) -> MetricType:
    return {
        f"{stage}_{key}" if not key.startswith(stage) else key: value
        for key, value in metrics.items()
    }


def flatten_train_val_metrics(
    train_metrics: ForwardOutputType,
    val_metrics: Optional[ForwardOutputType] = None,
    prepend_stage: bool = False,
) -> MetricType:
    flat_output: MetricType = dict()

    # need to distinguish between train and val metrics
    if val_metrics is not None:
        val_metrics = convert_output_to_dict(val_metrics)
        val_metrics = prepend_stage_to_metrics("val", val_metrics)

        prepend_stage = True
    else:
        val_metrics = dict()

    train_metrics = convert_output_to_dict(train_metrics)
    if prepend_stage:
        train_metrics = prepend_stage_to_metrics("train", train_metrics)

    flat_output.update(train_metrics)
    flat_output.update(val_metrics)

    return flat_output



from math import isinf, isnan
from typing import TYPE_CHECKING, Literal, Optional, cast

from lightning_cv.callbacks.base import Callback
from lightning_cv.callbacks.utils import remove_old_checkpoints

if TYPE_CHECKING:
    from lightning_cv.trainer import CrossValidationTrainer


class ModelCheckpoint(Callback):
    __checkpoint_glob__ = "epoch={epoch}_fold*"

    def __init__(
        self,
        save_last: bool = False,
        save_top_k: int = 1,
        every_n_epochs: Optional[int] = None,
        monitor: str = "loss",
        mode: Literal["min", "max"] = "min",
    ):
        self.save_last = save_last
        self.save_top_k = save_top_k
        self.every_n_epochs = every_n_epochs or 1
        self.monitor = monitor
        self.mode = mode

        # map epoch to val metric being monitored
        self.best_k_models: dict[int, float] = dict()

        self.filter_fn = max if self.mode == "min" else min

    def worst_of_best_epochs(self) -> int:
        if self.best_k_models:
            # filter out worst of best models
            # if minimizing value, then take max (worst) score of best models
            # if maximizing value, then take min (worst) score of best models
            return self.filter_fn(
                self.best_k_models, key=self.best_k_models.__getitem__
            )
        return -1

    def on_train_fold_end(self, trainer: "CrossValidationTrainer"):
        # need to check for this directly instead of using trainer property
        # if training stops during the first epoch, this attribute will not exist
        if hasattr(trainer, "_current_val_metrics"):
            # look at validation metrics at the end of "training" all folds
            # for fold in fold -> this does train, val, scheduler_step
            val_metric = cast(float, trainer._current_val_metrics[self.monitor].item())
            if trainer.current_epoch % self.every_n_epochs == 0:
                self._save_topk_ckpt(trainer, val_metric)

    def _save_topk_ckpt(self, trainer: "CrossValidationTrainer", metric: float):
        if self.save_top_k == 0:
            return

        k = len(self.best_k_models) + 1 if self.save_top_k == -1 else self.save_top_k

        if isnan(metric) or isinf(metric):
            # don't store, especially problematic with nan
            # since nan is not considered by min/max
            return

        # store current val loss
        self.best_k_models[trainer.current_epoch] = metric

        kth_best_model_epoch = self.worst_of_best_epochs()

        if len(self.best_k_models) == k + 1:
            del_epoch = kth_best_model_epoch
            if del_epoch != -1:  # pragma: no cover <- this is tested
                self.best_k_models.pop(del_epoch)
                glob = self.__checkpoint_glob__.format(epoch=del_epoch)

                remove_old_checkpoints(trainer.checkpoint_dir, glob)

            # if current epoch would be deleted immediately don't save it
            if del_epoch != trainer.current_epoch:
                self.save(trainer)
        else:
            # otherwise just save the checkpoint
            self.save(trainer)

    def on_train_end(self, trainer: "CrossValidationTrainer"):
        if self.save_last:
            self.save(trainer, suffix="last", include_val_loss=False)

    def save(self, trainer: "CrossValidationTrainer", **kwargs):
        if "include_val_loss" not in kwargs:
            kwargs["include_val_loss"] = True

        if not trainer.should_stop:
            # don't save if trainer was told to stop from external signals
            trainer.save(trainer.fold_manager, **kwargs)

    def __repr__(self) -> str:
        kwargs = dict(
            save_last=self.save_last,
            save_top_k=self.save_top_k,
            every_n_epochs=self.every_n_epochs,
            monitor=self.monitor,
            mode=self.mode,
        )
        kwargs_str = ", ".join(f"{key}={value}" for key, value in kwargs.items())
        return f"{self.__class__.__name__}({kwargs_str})"



from lightning.fabric.strategies.ddp import DDPStrategy
from lightning.fabric.strategies.deepspeed import DeepSpeedStrategy
from lightning.fabric.strategies.fsdp import FSDPStrategy
from lightning.fabric.strategies.strategy import Strategy
from lightning.fabric.strategies.xla import XLAStrategy

_DISTRIBUTED_STRATEGIES = (DDPStrategy, DeepSpeedStrategy, FSDPStrategy, XLAStrategy)


def is_distributed(strategy: Strategy) -> bool:
    if hasattr(strategy, "is_distributed"):  # pragma: no cover
        return strategy.is_distributed  # type: ignore

    return isinstance(strategy, _DISTRIBUTED_STRATEGIES)



import copy
from typing import Optional

import attrs

from lightning_cv.typehints import KwargType
from lightning_cv.utils.serde import config_deserializer


def _included_nested_dataclass(model, include: set[str]) -> dict[str, bool]:
    fields = attrs.fields(type(model))
    is_nested = {}
    for included_field in include:
        field: attrs.Attribute = getattr(fields, included_field)
        is_nested[included_field] = attrs.has(field.type)  # type: ignore[arg-type]

    return is_nested


def model_dump(
    model,
    include: Optional[set[str]] = None,
    exclude: Optional[set[str]] = None,
) -> KwargType:
    """Convert `attrs.define` dataclasses to a dictionary. Optionally include or exclude fields
    ONLY from the top-most level. Only `include` OR `exclude` can be used, not both.

    Args:
        model `attrs.define` dataclass instance
        include (Optional[Set[str]], optional): set of top-level keys to include.
            Defaults to None, meaning include ALL.
        exclude (Optional[Set[str]], optional): set of top-level keys to exclude.
            Defaults to None, meaning exclude NOTHING.

    Raises:
        ValueError: if model is not an `attrs.define` dataclass instance

    Returns:
        KwargType (dict[str, Any]): dictionary representation of the dataclass with all str
            keys
    """
    if not attrs.has(model):
        raise ValueError(f"model must be a `attrs.define` dataclass got {type(model)}")

    if include and exclude:
        raise ValueError("Only include OR exclude can be used, not both")

    asdict_kwargs = {}
    if include is not None:
        asdict_kwargs["filter"] = attrs.filters.include(*include)

        # handle special case for dumping nested models
        included_nested_dataclass = _included_nested_dataclass(model, include)
        if any(included_nested_dataclass.values()):
            # need to recurse? or just grab the entire thing?
            # TODO: this only works for one level of nesting
            ser = attrs.asdict(model, **asdict_kwargs)
            for key, is_nested in included_nested_dataclass.items():  # pragma: no cover
                if is_nested:
                    value = getattr(model, key)
                    ser[key] = model_dump(value)

            return ser

    elif exclude is not None:
        asdict_kwargs["filter"] = attrs.filters.exclude(*exclude)

    return attrs.asdict(model, **asdict_kwargs)


@attrs.define
class AttrsDataclassUtilitiesMixin:
    """Inherit from this for pure `attrs.define` utilities methods, including serialization
    (`.to_dict`), deserilization (`.from_dict`), and copying (`.copy`).

    Subclasses are expected to also be decorated with `attrs.define`.
    """

    def to_dict(
        self, include: Optional[set[str]] = None, exclude: Optional[set[str]] = None
    ):
        # for internal config objects in this lib, just use `attrs.asdict` which is a bit
        # nicer than what cattrs does
        return model_dump(self, include, exclude)

    @classmethod
    def from_dict(cls, data: KwargType):
        return config_deserializer(data, cls)

    def copy(self, deep: bool = False):
        return copy.deepcopy(self) if deep else copy.copy(self)



import logging

from lightning_cv.__metadata__ import (
    __author__,
    __description__,
    __license__,
    __title__,
    __version__,
)
from lightning_cv.config import CrossValidationTrainerConfig
from lightning_cv.data import CrossValidationDataModule, CrossValidationDataModuleMixin
from lightning_cv.split import BaseCrossValidator, BaseGroupCrossValidator
from lightning_cv.trainer import CrossValidationTrainer

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)
_logger.addHandler(logging.StreamHandler())

__all__ = [
    "CrossValidationDataModule",
    "CrossValidationDataModuleMixin",
    "BaseCrossValidator",
    "BaseGroupCrossValidator",
    "CrossValidationTrainer",
    "CrossValidationTrainerConfig",
    "__author__",
    "__description__",
    "__license__",
    "__title__",
    "__version__",
]



from collections.abc import Sequence
from pathlib import Path

from lightning_cv.callbacks.base import CallbackT, _CallbackT


def standardize_callbacks(callbacks: CallbackT) -> list[_CallbackT]:
    """Standardize all possible callbacks when creating a `CrossValidationTrainer` or
    `LightningFabric` instance into a list."""
    if callbacks is None:
        return []

    if isinstance(callbacks, Sequence):
        return list(callbacks)

    return [callbacks]


def remove_old_checkpoints(checkpoint_dir: Path, pattern: str = "*.ckpt"):
    """Remove old checkpoints from a directory based on a globbing pattern.

    Args:
        checkpoint_dir (Path): checkpointing directory
        pattern (Optional[str], optional): glob pattern for checkpoint files to remove. Defaults to '*.ckpt'.
    """
    for file in checkpoint_dir.glob(pattern):
        file.unlink()



from typing import TYPE_CHECKING

from torch.optim import Optimizer

from lightning_cv.callbacks.base import Callback
from lightning_cv.typehints import KwargType

if TYPE_CHECKING:
    from lightning_cv.trainer import CrossValidationTrainer


class LearningRateMonitor(Callback):
    def on_before_log_metrics(
        self,
        trainer: "CrossValidationTrainer",
        metrics: KwargType,
        optimizer: Optimizer,
    ):
        # TODO: make more complex
        lr = optimizer.param_groups[0]["lr"]
        metrics["lr"] = lr



"""(De)-serializing config objects for `LightningModule` and `LightningDataModule` construction.
"""

from enum import Enum

import cattrs

from lightning_cv._optional import HAS_PYDANTIC, PYDANTIC_V2
from lightning_cv.typehints import _T, KwargType
from lightning_cv.typehints.config import ConfigT

# TODO: might want to change enum handling?
# if users want to consider an enum as an hparam, then it may be challenging or awkward to accept user input
# which will most likely be the member name
# NOTE: users should create custom attrs.define converters that handle this instead....?
# could also create a custom converter for enums that converts to their values as a helpful utility
# TODO: make a test for the above case with enums for tuning tests
# the solution however will probably be to require a map for enums in the hparam config file that maps the name
# categorical choice to the enum value?
_CONVERTER = cattrs.Converter(prefer_attrib_converters=True)

# coverage has issues with thinking this block is not tested but it is
if HAS_PYDANTIC:  # pragma: no cover
    from pydantic import BaseModel

    @_CONVERTER.register_unstructure_hook
    def _pydantic_to_dict(value: BaseModel) -> KwargType:
        # mostly let pydantic handle it's own serialization, HOWEVER
        # pydantic does not serialize enums to their values like cattrs does,
        # so we need to do that manually
        ser: KwargType = value.model_dump() if PYDANTIC_V2 else value.dict()

        # walk through the dictionary and convert any Enum values to their values
        stack: list[KwargType] = [ser]
        while stack:
            curr = stack.pop()
            for key, value in curr.items():
                if isinstance(value, dict):
                    stack.append(value)
                elif isinstance(value, Enum):
                    curr[key] = value.value
        return ser

    @_CONVERTER.register_structure_hook
    def _pydantic_from_dict(value: KwargType, cls: type[BaseModel]) -> BaseModel:
        # pydantic can deserialize enum values back to enum members like cattrs
        # so don't need to worry about that
        if PYDANTIC_V2:
            return cls.model_validate(value)
        return cls.validate(value)


def config_serializer(config: ConfigT) -> KwargType:
    config_ser = _CONVERTER.unstructure(config)

    # TODO: may need to remove this depending on how we expect users to use configs?
    if not isinstance(config_ser, dict):
        raise ValueError(
            f"Could not automatically serialize the provided config: {config} of type "
            f"{type(config)}. It must be one of the following supported types: "
            "`dataclasses.dataclass`, `dict`, `pydantic.BaseModel`, or `attrs.define` dataclass."
        )

    return config_ser


def config_deserializer(serialized_config: KwargType, cls: type[_T]) -> _T:
    return _CONVERTER.structure(serialized_config, cls)



import json
from logging import getLogger
from math import isinf, isnan
from pathlib import Path
from typing import TYPE_CHECKING, cast

import optuna

from lightning_cv.callbacks.perf_monitor import ValPerfPlateauMonitor
from lightning_cv.utils.stopping import StopReasons

if TYPE_CHECKING:
    from lightning_cv.trainer import CrossValidationTrainer

logger = getLogger(__name__)


class TrialPruning(ValPerfPlateauMonitor):
    # this code is basically taken from the optuna pytorch lightning integration
    # and lightly modified to work with the custom CV Trainer in this module
    _EPOCH_KEY = "ddp_pl:epoch"
    _INTERMEDIATE_VALUE = "ddp_pl:intermediate_value"
    _PRUNED_KEY = "ddp_pl:pruned"
    _SEPARATOR = "*" * 15

    def __init__(
        self, trial: optuna.Trial, monitor: str = "loss", verbose: bool = False
    ):
        super().__init__(monitor=monitor)

        self._trial = trial
        self.is_ddp_backend = False
        self.verbose = verbose

    def on_train_start(self, trainer: "CrossValidationTrainer"):
        self.is_ddp_backend = trainer.is_distributed
        if self.is_ddp_backend:
            if trainer.is_global_zero:
                self._trial.storage.set_trial_system_attr(
                    self._trial._trial_id,
                    self._INTERMEDIATE_VALUE,
                    dict(),
                )

        logger.info(
            f"\n{self._SEPARATOR} Trial {self._trial._trial_id-1} BEGIN {self._SEPARATOR}"
            "\n"
        )

    def on_train_end(self, trainer: "CrossValidationTrainer"):
        logger.info(
            f"\n{self._SEPARATOR} Trial {self._trial._trial_id-1} END {self._SEPARATOR}\n"
        )

    def _distributed_check_prune(
        self,
        trainer: "CrossValidationTrainer",
        current_score: float,
        should_stop: bool = False,
    ):
        current_epoch = trainer.current_epoch
        if trainer.is_global_zero:
            should_stop = self._trial.should_prune()

            _trial_id = self._trial._trial_id
            _study = self._trial.study
            _trial_system_attrs = _study._storage.get_trial_system_attrs(_trial_id)

            intermediate_values = _trial_system_attrs.get(self._INTERMEDIATE_VALUE)
            intermediate_values[current_epoch] = current_score  # type: ignore[index]
            self._trial.storage.set_trial_system_attr(
                self._trial._trial_id,
                self._INTERMEDIATE_VALUE,
                intermediate_values,
            )

        # if ANY process has decided to stop
        should_stop = trainer.fabric.strategy.reduce_boolean_decision(
            should_stop, all=False
        )
        trainer.should_stop = trainer.should_stop or should_stop

        if not should_stop:
            return

        trainer.status = StopReasons.PRUNED_TRIAL

        if trainer.is_global_zero:
            # Update system_attr from global zero process.
            self._trial.storage.set_trial_system_attr(
                self._trial._trial_id, self._PRUNED_KEY, True
            )
            self._trial.storage.set_trial_system_attr(
                self._trial._trial_id, self._EPOCH_KEY, current_epoch
            )

    def on_train_fold_end(self, trainer: "CrossValidationTrainer"):
        current_score = trainer.current_val_metrics[self.monitor].item()
        current_epoch = trainer.current_epoch
        should_stop = False
        stop_reason = StopReasons.PRUNED_TRIAL
        if isnan(current_score) or isinf(current_score):
            # stop if we start seeing nan/inf
            should_stop = True
            stop_reason = StopReasons.LOSS_NAN_OR_INF

        # stores current_score into the trial's intermediate values
        self._trial.report(current_score, current_epoch)

        # single process training
        if not self.is_ddp_backend:
            if not self._trial.should_prune() and not should_stop:
                # both the trial must think it doesn't need to prune
                # AND an external factors, like the occurence of nan loss
                return
            trainer.status = stop_reason
            raise optuna.TrialPruned(self.prune_message(epoch=current_epoch))

        # distributed training
        self._distributed_check_prune(trainer, current_score, should_stop)

    def on_validation_end_per_fold(self, trainer: "CrossValidationTrainer"):
        super().on_validation_end_per_fold(trainer)

        if (
            trainer.should_stop
            and trainer.current_epoch < self.min_epochs
            and trainer.status != StopReasons.TIMEOUT
        ):
            # if we have trained a sufficient number of epochs just report the value
            # otherwise prune the trial
            # also don't prune for a timeout, which would stop the trainer
            raise optuna.TrialPruned(self.prune_message(epoch=trainer.current_epoch))

            # immediately after this, the current_val_metrics will be calculated and reported

    def check_pruned(self):
        """Raise :class:`optuna.TrialPruned` manually if pruned.

        Currently, ``intermediate_values`` are not properly propagated between processes
        due to storage cache. Therefore, necessary information is kept in
        trial_system_attrs when the trial runs in a distributed situation. Please call
        this method right after calling
        ``lightning_cv.CrossValidationTrainer.train_with_cross_validation()``.

        If a callback doesn't have any backend storage for DDP, this method does
        nothing.
        """
        _trial_id = self._trial._trial_id
        _study = self._trial.study
        # Confirm if storage is not InMemory in case this method is called in a
        # non-distributed situation by mistake.
        if not isinstance(_study._storage, optuna.storages._CachedStorage):
            return

        _trial_system_attrs = _study._storage._backend.get_trial_system_attrs(_trial_id)
        is_pruned = _trial_system_attrs.get(self._PRUNED_KEY)
        intermediate_values = _trial_system_attrs.get(self._INTERMEDIATE_VALUE)

        # Confirm if DDP backend is used in case this method is called from a
        # non-DDP situation by mistake.
        if intermediate_values is None:
            return
        for epoch, score in intermediate_values.items():
            self._trial.report(score, step=int(epoch))
        if is_pruned:
            epoch = cast(int, _trial_system_attrs.get(self._EPOCH_KEY))
            raise optuna.TrialPruned(self.prune_message(epoch=epoch))

    def prune_message(self, epoch: int):
        msg = [f"Trial was pruned at epoch {epoch}"]

        if self.verbose:
            msg.append(f"with parameters {self._trial.params}")

        return " ".join(msg)


class OptunaHyperparameterLogger:
    def __init__(
        self,
        root_dir: Path,
        expt_name: str,
        logfile: str = "sampled_hparams.json",
        monitor: str = "loss",
    ):
        self.root_dir = root_dir
        self.expt_name = expt_name
        self.logfile = logfile
        self.monitor = monitor if monitor.startswith("val_") else f"val_{monitor}"

    def __call__(self, study: optuna.Study, trial: optuna.trial.FrozenTrial):
        trial_id = (
            trial.number
        )  # TODO: this no longer matches up with the logger version
        logdir = self.root_dir / self.expt_name / f"version_{trial_id}"
        logfile = logdir.joinpath(self.logfile)

        record = {
            "monitor": self.monitor,
            "performance": trial.intermediate_values,
            "hparams": trial.params,
        }

        with logfile.open("w") as fp:
            json.dump(record, fp, indent=4)
            fp.write("\n")



from statistics import mean, stdev
from typing import TYPE_CHECKING, Optional

from lightning_cv.callbacks.base import Callback
from lightning_cv.typehints import MetricType
from lightning_cv.utils.stopping import StopReasons

if TYPE_CHECKING:
    from lightning import LightningModule

    from lightning_cv.trainer import CrossValidationTrainer


class ValPerfPlateauMonitor(Callback):
    def __init__(
        self,
        monitor="loss",
        min_metric: float = 1e-3,
        threshold_std: float = 1e-6,
        min_epochs: int = 3,
    ):
        self.monitor = monitor
        self.min_metric = min_metric
        self.threshold_std = threshold_std
        self.min_epochs = min_epochs

        self.metrics: list[float] = list()

    def mean(self) -> float:
        return mean(self.metrics)

    def std(self, xbar: Optional[float] = None) -> float:
        return stdev(self.metrics, xbar=xbar)

    def push_metric(self, output: MetricType):
        metric = output[self.monitor].item()
        self.metrics.append(metric)

    def clear_metrics(self):
        self.metrics.clear()

    def should_stop(self) -> bool:
        if len(self.metrics) <= 1:
            # likely due to timeout or some other external stop
            # let that callback handle chaning trainer status
            return False

        mean = self.mean()
        std = self.std(xbar=mean)

        # if the std of metric is near 0.0, suggests learning has stopped
        # if the mean is greater than the min_metric, then learning was not optimal since
        # the metric has not approached the min_metric
        return std <= self.threshold_std and mean > self.min_metric

    def report(self, trainer: "CrossValidationTrainer"):
        should_stop = self.should_stop()

        if should_stop:
            if trainer.is_distributed:  # pragma: no cover
                should_stop = trainer.fabric.strategy.reduce_boolean_decision(
                    should_stop, all=False
                )
            trainer.should_stop = trainer.should_stop or should_stop
            trainer.status = StopReasons.PERFORMANCE_STALLED

    def on_validation_start_per_fold(
        self, trainer: "CrossValidationTrainer", model: "LightningModule", total: int
    ):
        self.clear_metrics()

    def on_validation_batch_end_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        output: MetricType,
        batch,
        batch_idx: int,
    ):
        self.push_metric(output)

    def on_validation_end_per_fold(self, trainer: "CrossValidationTrainer"):
        self.report(trainer)



from lightning_cv.callbacks.base import Callback
from lightning_cv.callbacks.checkpoint import ModelCheckpoint
from lightning_cv.callbacks.lr_monitor import LearningRateMonitor
from lightning_cv.callbacks.perf_monitor import ValPerfPlateauMonitor
from lightning_cv.callbacks.stopping import EarlyStopping
from lightning_cv.callbacks.timer import Timer

__all__ = [
    "Callback",
    "ModelCheckpoint",
    "LearningRateMonitor",
    "ValPerfPlateauMonitor",
    "EarlyStopping",
    "Timer",
]



import sys
from importlib.util import find_spec
from typing import TYPE_CHECKING, Literal, Optional, Union, cast

from lightning_utilities.core.apply_func import apply_to_collection
from torch import Tensor

from lightning_cv.callbacks.base import Callback
from lightning_cv.typehints import MetricType

if TYPE_CHECKING:
    from lightning import LightningModule

    from lightning_cv.trainer import CrossValidationTrainer


if find_spec("ipywidgets") is not None:  # pragma: no cover
    from tqdm.auto import tqdm as _tqdm
else:  # pragma: no cover
    from tqdm import tqdm as _tqdm


_PAD_SIZE = 5

ExtractedMetricType = dict[str, float]


class Tqdm(_tqdm):  # type: ignore
    @staticmethod
    def format_num(n: Union[int, float, str]) -> str:
        """Add additional padding to the formatted numbers."""
        should_be_padded = isinstance(n, (float, str))
        if not isinstance(n, str):
            n = _tqdm.format_num(n)
            assert isinstance(n, str)
        if should_be_padded and "e" not in n:
            if "." not in n and len(n) < _PAD_SIZE:
                try:
                    _ = float(n)
                except ValueError:
                    return n
                n += "."
            n += "0" * (_PAD_SIZE - len(n))
        return n


class TqdmProgressBar(Callback):
    def init_progbar(self, stage: Literal["train", "validate"], **kwargs):
        progbar = Tqdm(
            dynamic_ncols=True,
            leave=stage != "validate",
            smoothing=0.0,
            file=sys.stdout,
            **kwargs,
        )

        return progbar

    def on_train_start(self, trainer: "CrossValidationTrainer"):
        self.train_pbar = self.init_progbar(stage="train")

    def on_train_epoch_start_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        model: "LightningModule",
        total: int,
        current_epoch: int,
        current_fold: int,
    ):
        # update total
        self.train_pbar.reset(total=total)
        # reset batch number to 0
        self.train_pbar.initial = 0
        self.current_epoch = current_epoch
        self.train_pbar.set_description(f"Epoch {current_epoch} Fold {current_fold}")

    def on_train_batch_end_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        output: MetricType,
        batch,
        batch_idx: int,
    ):
        n = batch_idx + 1
        _update_pbar(self.train_pbar, value=n)
        self._train_log(output)

    def on_train_epoch_end_per_fold(
        self, trainer: "CrossValidationTrainer", output: MetricType
    ):
        self._train_log(output)

    def _train_log(self, output: MetricType):
        prefix = "train" if self.current_epoch == 0 else None
        metrics = self.get_metrics(output)
        _set_postfix(self.train_pbar, metrics, prefix)

    def on_train_end(self, trainer: "CrossValidationTrainer"):
        self.train_pbar.close()

    def on_validation_start_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        model: "LightningModule",
        total: int,
    ):
        self.val_pbar = self.init_progbar(stage="validate", total=total)
        self.val_pbar.set_description(
            f"Epoch {self.current_epoch} Fold {trainer.current_fold} validation"
        )

    def on_validation_batch_end_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        output: MetricType,
        batch,
        batch_idx: int,
    ):
        n = batch_idx + 1
        _update_pbar(self.val_pbar, value=n)

    def on_validation_end_per_fold(self, trainer: "CrossValidationTrainer"):
        self.val_pbar.close()

    def get_metrics(self, candidates: MetricType) -> ExtractedMetricType:
        float_candidates = cast(
            ExtractedMetricType,
            apply_to_collection(data=candidates, dtype=Tensor, function=_extract_value),
        )

        return float_candidates


def _extract_value(tensor: Tensor):
    return tensor.item()


def _update_pbar(pbar: Tqdm, value: int):
    if (
        not pbar.disable
    ):  # pragma: no cover <- doesn't need to be tested since this is a tqdm method
        pbar.n = value
        pbar.refresh()


def _set_postfix(
    pbar: Tqdm,
    metric: ExtractedMetricType,
    prefix: Optional[str] = None,
):
    postfix = dict()
    if prefix is None:
        # don't append a prefix name like train/val
        prefix = "{key}"
    else:
        prefix = prefix.strip() + "_{key}"

    for key, value in metric.items():
        postfix[prefix.format(key=key)] = f"{value:.3f}"

    if (
        postfix
    ):  # pragma: no cover <- doesn't need to be tested since this is a tqdm method
        pbar.set_postfix(postfix)



from importlib.util import find_spec

HAS_PYDANTIC = find_spec("pydantic") is not None
HAS_OPTUNA = find_spec("optuna") is not None

if HAS_PYDANTIC:
    import pydantic

    PYDANTIC_V2 = int(pydantic.version.VERSION.split(".")[0]) == 2

    def is_pydantic_basemodel(obj) -> bool:
        """Checks if the given input is an instance of a Pydantic BaseModel OR a subclass of it."""
        if isinstance(obj, type):
            return issubclass(obj, pydantic.BaseModel)
        return isinstance(obj, pydantic.BaseModel)

else:  # pragma: no cover <- there is no need to test this...
    PYDANTIC_V2 = False

    def is_pydantic_basemodel(obj) -> bool:
        return False



import dataclasses
from collections import defaultdict
from collections.abc import Mapping
from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
from inspect import Signature, getfile, signature
from pathlib import Path
from typing import Any, Optional, TypedDict, Union
from warnings import warn

import attrs
from lightning import Fabric, LightningModule
from lightning.fabric.wrappers import _unwrap_objects as fabric_unwrap
from torch.optim import Optimizer

from lightning_cv._optional import is_pydantic_basemodel
from lightning_cv.config import FoldState
from lightning_cv.typehints import FilePath, KwargType, MetricType
from lightning_cv.utils.serde import config_deserializer, config_serializer

MODEL_CLASS_KEY = "MODEL_CLASS"
OPTIMIZER_CLASS_KEY = "OPTIMIZER_CLASS"
MODEL_INIT_DESERIALIZABLE_KEY = "MODEL_INIT_DESERIALIZABLE"
MODEL_INIT_ARGS_KEY = "MODEL_INIT_ARGS"
MODEL_STATE_DICT = "model"
OPTIMIZER_STATE_DICT = "optimizer"
SCHEDULER_CLASS_KEY = "SCHEDULER_CLASS"
SCHEDULER_STATE_DICT = "scheduler"

_MISSING_SENTINEL = object()


class _ClassImportMetadata(TypedDict):
    module: str
    classname: str
    module_path: str


def _get_class_import_path(obj) -> _ClassImportMetadata:
    module = obj.__module__
    if isinstance(obj, type):  # passed class
        classname = obj.__name__
        module_path = getfile(obj)
    else:  # passed instance
        classname = obj.__class__.__name__
        module_path = getfile(obj.__class__)

    return _ClassImportMetadata(
        module=module, classname=classname, module_path=module_path
    )


def _is_deserializable(T: type) -> bool:
    return attrs.has(T) or dataclasses.is_dataclass(T) or is_pydantic_basemodel(T)


def _model_init_signature(model: type[LightningModule]) -> Signature:
    return signature(model)


def _check_model_init_signature_for_deserialization(
    sig: Signature,
) -> list[tuple[str, _ClassImportMetadata]]:
    deserializable_init_args: list[tuple[str, _ClassImportMetadata]] = list()

    for key, param in sig.parameters.items():
        if param.annotation is not param.empty and _is_deserializable(param.annotation):
            class_path = _get_class_import_path(param.annotation)
            deserializable_init_args.append((key, class_path))

    return deserializable_init_args


class CheckpointLoaderMixin:
    current_epoch: int
    global_step_per_fold: dict[int, int]
    is_first_epoch: bool
    current_val_metrics: MetricType
    checkpoint_dir: Path
    fabric: Fabric

    @staticmethod
    def _try_load_module_from_filepath(module_name: str, module_path: str):
        # this is the recipe from python stdlib docs
        import sys

        spec = spec_from_file_location(module_name, module_path)
        module = module_from_spec(spec)  # type: ignore
        sys.modules[module_name] = module
        spec.loader.exec_module(module)  # type: ignore
        return module

    @staticmethod
    def _try_load_module(module_name: str, module_path: str):
        try:
            return import_module(module_name)
        except ModuleNotFoundError:
            try:
                return CheckpointLoaderMixin._try_load_module_from_filepath(
                    module_name, module_path
                )
            except Exception as e:
                raise ImportError(f"Could not import module {module_path}") from e

    @staticmethod
    def _try_import_class(import_metadata: _ClassImportMetadata):
        module = CheckpointLoaderMixin._try_load_module(
            import_metadata["module"], import_metadata["module_path"]
        )
        try:
            return getattr(module, import_metadata["classname"])
        except AttributeError as e:
            raise ImportError(
                f"Could not import class {import_metadata['classname']} from module {module}: {import_metadata}"
            ) from e

    def _set_self_state(self, state_dict: KwargType):
        self.current_fold = state_dict.pop("current_fold")
        self.current_epoch = state_dict.pop("current_epoch")
        self.global_step_per_fold[self.current_fold] = state_dict.pop("global_step")

    def _load_model(self, state_dict: KwargType) -> LightningModule:
        model_class_path = state_dict.pop(MODEL_CLASS_KEY)
        model_class = self._try_import_class(model_class_path)

        if not issubclass(model_class, LightningModule):
            raise ValueError(
                f"Model class {model_class} is not a subclass of LightningModule"
            )

        model_init_args: KwargType = state_dict.pop(MODEL_INIT_ARGS_KEY)

        # need to deserialize any deserializable init args
        deserializable_init_args = state_dict.pop(MODEL_INIT_DESERIALIZABLE_KEY)

        for init_arg, init_arg_class_path in deserializable_init_args:
            init_arg_class = self._try_import_class(init_arg_class_path)
            config = model_init_args.pop(init_arg)
            model_init_args[init_arg] = config_deserializer(config, init_arg_class)

        # then build model
        model = model_class(**model_init_args)
        model.load_state_dict(state_dict.pop(MODEL_STATE_DICT))

        return model

    def _load_optimizer(
        self, state_dict: KwargType, model: LightningModule
    ) -> Optimizer:
        optimizer_class_path = state_dict.pop(OPTIMIZER_CLASS_KEY)
        optimizer_class = self._try_import_class(optimizer_class_path)

        optimizer = optimizer_class(
            model.parameters()
        )  # TODO: need to capture init args from state_dict
        optimizer.load_state_dict(state_dict.pop(OPTIMIZER_STATE_DICT))

        return optimizer

    def _load_scheduler(self, state_dict: KwargType, optimizer: Optimizer):
        scheduler_class_path = state_dict.pop(SCHEDULER_CLASS_KEY)
        scheduler_class = self._try_import_class(scheduler_class_path)

        scheduler = scheduler_class(
            optimizer
        )  # TODO: need to capture init args from state_dict
        scheduler.load_state_dict(state_dict.pop(SCHEDULER_STATE_DICT))

        return scheduler

    def load(self, checkpoint_path: FilePath) -> FoldState:
        warn(
            "Support for loading checkpoints is beta and may not work as expected",
            UserWarning,
        )
        state_dict = self.fabric.load(checkpoint_path)

        self._set_self_state(state_dict)
        model = self._load_model(state_dict)
        optimizer = self._load_optimizer(state_dict, model)
        if state_dict[SCHEDULER_STATE_DICT] is not None:
            scheduler = self._load_scheduler(state_dict, optimizer)
        else:
            scheduler = None

        return FoldState(model=model, optimizer=optimizer, scheduler=scheduler)

    def load_all_folds(
        self, checkpoint_dir: Path, ext: str = "ckpt"
    ) -> dict[int, FoldState]:
        paths = list(checkpoint_dir.glob(f"*.{ext}"))
        ckpt_status: defaultdict[str, list[Path]] = defaultdict(list)
        for path in paths:
            name = path.name
            if "training_complete" in name:
                ckpt_status["timeout"].append(path)
            elif "last" in name:
                ckpt_status["last"].append(path)
            else:
                ckpt_status["default"].append(path)

        fold_manager: dict[int, FoldState] = dict()
        if ckpt_status["timeout"]:
            for path in ckpt_status["timeout"]:
                # TODO: load checkpoints AND finish remaining checkpoints
                pass

        for path in paths:
            # this should set self.current_fold
            state = self.load(path)
            fold_manager[self.current_fold] = state
        return fold_manager


class CheckpointSaverMixin:
    current_epoch: int
    global_step_per_fold: dict[int, int]
    is_first_epoch: bool
    current_val_metrics: MetricType
    checkpoint_dir: Path
    fabric: Fabric

    def _unwrap_fabric_objects_in_place(self, state: FoldState):
        for key in ["model", "optimizer"]:
            setattr(state, key, fabric_unwrap(getattr(state, key)))

    def _checkpoint_name(
        self,
        fold: int,
        suffix: Optional[str] = None,
        include_val_loss: bool = True,
        ext: str = "ckpt",
        **name_kwargs,
    ) -> Path:
        # use new dict for consistent key order
        _name_kwargs = dict()
        _name_kwargs["epoch"] = self.current_epoch
        _name_kwargs["fold"] = fold

        if include_val_loss and not self.is_first_epoch:
            _name_kwargs["loss"] = f"{self.current_val_metrics['loss']:.4f}"
        else:
            _name_kwargs.update(name_kwargs)

        if suffix is not None:
            _name_kwargs["suffix"] = suffix

        _name = "_".join(
            f"{k}={v}" if k != "suffix" else str(v) for k, v in _name_kwargs.items()
        )
        return self.checkpoint_dir / f"{_name}.{ext}"

    def _serialize_state_for_saving(self, state: FoldState, fold: int) -> KwargType:
        self._unwrap_fabric_objects_in_place(state)

        ser_state = state.to_dict()
        ser_state.update(
            global_step=self.global_step_per_fold[fold],
            current_epoch=self.current_epoch,
            current_fold=fold,
        )

        # need to keep track of the optimizer and model classes
        # NOTE: this is required since lightning.Fabric does not allow module and optimizer class
        # types to be saved, even though this would probably be possible with pickling
        ser_state[MODEL_CLASS_KEY] = _get_class_import_path(state.model)
        ser_state[OPTIMIZER_CLASS_KEY] = _get_class_import_path(state.optimizer)
        if state.scheduler is not None:
            warn("Saving schedulers is beta and may not work as expected", UserWarning)
            ser_state[SCHEDULER_CLASS_KEY] = _get_class_import_path(
                state.scheduler["scheduler"]
            )

        # also add the init args for the model when reloading
        model_init_sig = _model_init_signature(state.model.__class__)  # type: ignore[arg-type]

        # NOTE: requires users to set all init args to the correspondingly named keys on the model
        model_init_args: KwargType = dict()
        for key in model_init_sig.parameters.keys():
            value = state.model.hparams.get(key, _MISSING_SENTINEL)

            if value is _MISSING_SENTINEL:
                value = getattr(state.model, key, _MISSING_SENTINEL)

            if value is _MISSING_SENTINEL:
                raise AttributeError(
                    f"Could not find attribute {key} in model {state.model}. This attribute "
                    "must be set as an instance attribute (ie self.key = value) in the model OR "
                    "be stored in the LightningModule's hparams dict (as a result of using "
                    "the LightningModule.save_hyperparameters method)."
                )

            model_init_args[key] = value

        ser_state[MODEL_INIT_ARGS_KEY] = config_serializer(model_init_args)

        # check if any init args are deserializable from dicts
        deserializable_model_init_args = (
            _check_model_init_signature_for_deserialization(model_init_sig)
        )

        ser_state[MODEL_INIT_DESERIALIZABLE_KEY] = deserializable_model_init_args

        return ser_state

    def _save(
        self,
        state: FoldState,
        fold: int,
        suffix: Optional[str] = None,
        ext: str = "ckpt",
        include_val_loss: bool = True,
        **name_kwargs: Any,
    ):
        ser_state = self._serialize_state_for_saving(state, fold)

        output = self._checkpoint_name(
            fold=fold,
            suffix=suffix,
            include_val_loss=include_val_loss,
            ext=ext,
            **name_kwargs,
        )

        self.fabric.save(output, ser_state)

    def save(
        self,
        states: Union[FoldState, Mapping[int, FoldState]],
        suffix: Optional[str] = None,
        ext: str = "ckpt",
        include_val_loss: bool = False,
        **name_kwargs,
    ):
        """Saves a checkpoint to the ``checkpoint_dir``

        Args:
            state: A mapping containing model, optimizer and lr scheduler.
        """
        if isinstance(states, Mapping):
            for fold, state in states.items():
                self._save(
                    state=state,
                    fold=fold,
                    suffix=suffix,
                    ext=ext,
                    include_val_loss=include_val_loss,
                    **name_kwargs,
                )
        else:
            self._save(
                state=states,
                suffix=suffix,
                ext=ext,
                include_val_loss=include_val_loss,
                **name_kwargs,
            )


class CurrentCheckpointsMixin:
    checkpoint_dir: Path

    def _keep_latest_checkpoints(self, paths: list[Path]) -> list[Path]:
        import re
        from operator import itemgetter

        pattern = re.compile(r"epoch=(\d+)_fold=(\d+)")

        path2ef = {
            path: tuple(int(x) for x in match.groups())
            for path in paths
            if (match := pattern.search(path.name))
        }
        latest_epoch = max(path2ef.values(), key=itemgetter(0))[0]
        latest_path2ef = {
            path: (epoch, fold)
            for path, (epoch, fold) in path2ef.items()
            if epoch == latest_epoch
        }

        return sorted(latest_path2ef, key=latest_path2ef.__getitem__)

    @staticmethod
    def get_latest_checkpoint(ckpt_dir: Path) -> Optional[Path]:
        """Returns the latest checkpoint from the ``checkpoint_dir``

        Args:
            checkpoint_dir: the directory to search for checkpoints
        """
        import os

        if not ckpt_dir.is_dir():
            return None

        # TODO: I don't think this works as expected
        items = sorted(os.listdir(ckpt_dir))

        if not items:
            return None

        return ckpt_dir.joinpath(items[-1])

    def current_ckptfiles(self) -> list[Path]:
        return list(self.checkpoint_dir.glob("*.ckpt"))


class CheckpointingMixin(
    CheckpointSaverMixin, CheckpointLoaderMixin, CurrentCheckpointsMixin
):
    pass



from typing import TYPE_CHECKING

from lightning_cv.callbacks.base import Callback

if TYPE_CHECKING:
    from lightning_cv.trainer import CrossValidationTrainer


class LogFlusher(Callback):
    def on_train_end(self, trainer: "CrossValidationTrainer"):
        # for CSVLogger, need to flush buffer at the end
        trainer.fabric.logger.save()



from typing import Callable, Iterator, Protocol

from numpy import int64
from numpy.typing import NDArray

Int64Array = NDArray[int64]
CVIterator = Iterator[tuple[Int64Array, Int64Array]]


class CrossValidator(Protocol):
    # ignore input types and just require to implement these methods
    split: Callable[..., CVIterator]



from pathlib import Path
from typing import Literal, Optional, Union

from attrs import define, field
from lightning import LightningModule
from lightning.fabric.wrappers import _FabricModule
from torch.optim import Optimizer

from lightning_cv.fabric import (
    AcceleratorT,
    CallbackT,
    LoggerT,
    PluginsT,
    PrecisionT,
    StrategyT,
)
from lightning_cv.typehints import Number, SchedulerConfigT
from lightning_cv.utils.dataclasses import AttrsDataclassUtilitiesMixin
from lightning_cv.utils.validators import (
    limit_batches_validator,
    non_negative_float,
    non_negative_int,
    positive_int,
)


@define
class FoldState(AttrsDataclassUtilitiesMixin):
    model: Union[LightningModule, _FabricModule]
    optimizer: Optimizer
    scheduler: Optional[SchedulerConfigT] = None


GradClipAlgorithms = Literal["norm", "value"]


@define
class CrossValidationTrainerConfig(AttrsDataclassUtilitiesMixin):
    accelerator: AcceleratorT = "auto"
    strategy: StrategyT = "auto"
    devices: Union[int, str, list[int]] = "auto"
    precision: PrecisionT = "32"
    plugins: PluginsT = None
    callbacks: CallbackT = None
    loggers: LoggerT = None
    max_epochs: int = field(default=1000, validator=positive_int)
    grad_accum_steps: int = field(default=1, validator=positive_int)
    gradient_clip_algorithm: GradClipAlgorithms = "norm"
    gradient_clip_val: float = field(default=0.0, validator=non_negative_float)
    limit_train_batches: Number = field(
        default=1.0,
        validator=limit_batches_validator,  # type: ignore
    )
    limit_val_batches: Number = field(
        default=1.0,
        validator=limit_batches_validator,  # type: ignore
    )
    validation_frequency: int = field(default=1, validator=non_negative_int)
    use_distributed_sampler: bool = True
    checkpoint_dir: Path = field(
        default=Path.cwd().joinpath("checkpoints"), converter=Path
    )
    checkpoint_frequency: int = field(default=1, validator=positive_int)
    monitor: str = field(
        default="val_loss", converter=lambda x: x if x.startswith("val") else f"val_{x}"
    )



from lightning_cv.tuning.callbacks import OptunaHyperparameterLogger, TrialPruning
from lightning_cv.tuning.hyperparameters import HparamRegistry
from lightning_cv.tuning.tuner import Tuner

__all__ = [
    "TrialPruning",
    "OptunaHyperparameterLogger",
    "HparamRegistry",
    "Tuner",
]



import sys
from typing import TYPE_CHECKING, cast

from lightning import LightningModule
from lightning.pytorch.utilities.model_summary.model_summary import summarize

from lightning_cv.callbacks.base import Callback

if TYPE_CHECKING:
    from lightning_cv.trainer import CrossValidationTrainer


class ModelSummary(Callback):
    def __init__(self, max_depth: int = 1):
        self.max_depth = max_depth

    def on_train_start(self, trainer: "CrossValidationTrainer"):
        model = cast(LightningModule, trainer.fold_manager[0].model)
        summary = summarize(model, max_depth=self.max_depth)
        sys.stderr.write(f"{repr(summary)}\n")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(max_depth={self.max_depth})"



from typing import TYPE_CHECKING

from lightning_cv.callbacks.base import Callback

if TYPE_CHECKING:
    from lightning import LightningModule

    from lightning_cv.trainer import CrossValidationTrainer


class TrainMode(Callback):
    def on_train_epoch_start_per_fold(
        self,
        trainer: "CrossValidationTrainer",
        model: "LightningModule",
        total: int,
        current_epoch: int,
        current_fold: int,
    ):
        model.train()


class EvalMode(Callback):
    def on_validation_start_per_fold(
        self, trainer: "CrossValidationTrainer", model: "LightningModule", total: int
    ):
        model.eval()



from abc import ABCMeta, abstractmethod

import numpy as np

from lightning_cv.typehints.split import CVIterator, Int64Array


class BaseCrossValidator(metaclass=ABCMeta):
    """Simpler CV iterface than sklearn. Only `split` and `get_n_splits` methods are required.

    The `split` method should yield a tuple of train and validation index arrays. It also takes
    no arguments, so it is recommended for the needed data to be passed in during initialization.
    """

    def get_X_index(self, X: Int64Array) -> Int64Array:
        n = X.shape[0]
        return np.arange(n)

    @abstractmethod
    def split(self) -> CVIterator: ...

    @abstractmethod
    def get_n_splits(self) -> int: ...


class BaseGroupCrossValidator(BaseCrossValidator):

    @property
    @abstractmethod
    def train_group_ids(self) -> list[int]: ...

    @property
    @abstractmethod
    def val_group_id(self) -> int: ...


class ImbalancedLeaveOneGroupOut(BaseGroupCrossValidator):
    """Group-based CV splitter. It is a variant of LeaveOneGroupOut where the validation set
    does not contain the most frequent group. The usecase is for highly imbalanced datasets
    in which the largest group substantially more frequent than the rest of the groups.
    """

    def __init__(self, groups: Int64Array) -> None:
        # only keep track of data indices for each fold
        self.X_idx = self.get_X_index(groups)

        self._val_group_id = -1
        self._train_group_ids: list[int] = list()

        self.uniq_groups: Int64Array
        self.groups: Int64Array
        # self.groups maps actual group labels to consecutively increasing,
        # ie [0, 1, ..., n-1, n]
        self.uniq_groups, self.groups = np.unique(groups, return_inverse=True)
        self.group_counts = np.bincount(self.groups)
        self._sort_groups()

        # don't have a fold for the most frequent group
        self.n_folds = self.uniq_groups.shape[0] - 1
        self.largest_group_id = np.argmax(self.group_counts)

    @property
    def train_group_ids(self) -> list[int]:
        return self._train_group_ids

    @train_group_ids.setter
    def train_group_ids(self, value: list[int]):
        self._train_group_ids = value

    @property
    def val_group_id(self) -> int:
        return self._val_group_id

    @val_group_id.setter
    def val_group_id(self, value: int):
        self._val_group_id = value

    def _sort_groups(self):
        # reverse sort, ie 0th element is count of most frequent group
        sort_idx = np.argsort(self.group_counts)[::-1]
        self.uniq_groups = self.uniq_groups[sort_idx]
        self.group_counts = self.group_counts[sort_idx]

    def __len__(self) -> int:
        return self.n_folds

    def get_n_splits(self) -> int:
        return self.n_folds

    def split(self) -> CVIterator:
        for fold_idx, group_id in enumerate(self.uniq_groups):
            if fold_idx == 0:
                # the first group will be the most frequent and
                # we don't want a fold where the most frequent group
                # is the validation set
                continue

            # keep track of current CV group id and the training groups
            self.val_group_id = group_id
            self.train_group_ids = self.uniq_groups[
                np.where(self.uniq_groups != group_id)
            ].tolist()

            val_mask = np.where(self.groups == group_id, True, False)
            train_mask = ~val_mask

            train_idx = self.X_idx[train_mask]
            val_idx = self.X_idx[val_mask]

            yield train_idx, val_idx



from importlib import metadata

package = __package__ or "lightning-cv"

_data = metadata.metadata(package)
__title__: str = _data["name"]
__description__: str = _data["summary"]
__version__: str = _data["version"]
__author__: str = _data["author"]
__license__: str = _data["license"]



from typing import TYPE_CHECKING, Any, Literal, Union

from lightning import Fabric as LightningFabric
from lightning_utilities.core.rank_zero import rank_zero_warn

if TYPE_CHECKING:
    from lightning.fabric.accelerators.accelerator import Accelerator
    from lightning.fabric.loggers.logger import Logger
    from lightning.fabric.plugins import CheckpointIO, ClusterEnvironment, Precision
    from lightning.fabric.strategies.strategy import Strategy
    from torch.nn import Module
    from torch.optim import Optimizer

from lightning import Callback as LightingCallback
from lightning import LightningModule

from lightning_cv.callbacks.base import Callback as LCVCallback
from lightning_cv.callbacks.base import CallbackT

PrecisionT = Union[
    None,
    Literal[
        64,
        32,
        16,
        "transformer-engine",
        "transformer-engine-float16",
        "16-true",
        "16-mixed",
        "bf16-true",
        "bf16-mixed",
        "32-true",
        "64-true",
        "64",
        "32",
        "16",
        "bf16",
    ],
]

_PluginsT = Union["Precision", "ClusterEnvironment", "CheckpointIO"]
PluginsT = Union[_PluginsT, list[_PluginsT], None]

AcceleratorT = Union[Literal["auto", "cpu", "gpu", "tpu", "cuda", "mps"], "Accelerator"]
StrategyT = Union[
    Literal["dp", "ddp", "ddp_spawn", "ddp_notebook", "fsdp", "deepspeed", "auto"],
    "Strategy",
]
DevicesT = Union[list[int], Literal["auto"], int]


LoggerT = Union["Logger", list["Logger"], None]


LightningFabric()


class Fabric(LightningFabric):

    _lcv_callbacks: list[LCVCallback]
    _model_callbacks: list[LightningModule]
    _lightning_callbacks: list[LightingCallback]

    def __init__(
        self,
        *,
        accelerator: AcceleratorT = "auto",
        strategy: StrategyT = "auto",
        devices: DevicesT = "auto",
        num_nodes: int = 1,
        precision: PrecisionT = None,
        plugins: PluginsT = None,
        callbacks: CallbackT = None,
        loggers: LoggerT = None,
    ) -> None:
        super().__init__(
            accelerator=accelerator,
            strategy=strategy,
            devices=devices,
            num_nodes=num_nodes,
            precision=precision,
            plugins=plugins,
            callbacks=callbacks,
            loggers=loggers,
        )
        # could probably merge lightning and model callbacks since they have the same callbacks
        # except I think they expect a lightning Trainer?
        self._lcv_callbacks = []
        self._model_callbacks = []
        self._lightning_callbacks = []
        # only move each callback once
        self._seen_callbacks = set()

        self._already_warned_about_callbacks = False

        self._move_all_callbacks()

    def _move_callback(self, callback):
        if callback in self._seen_callbacks:
            return

        if isinstance(callback, LCVCallback):
            self._lcv_callbacks.append(callback)
        elif isinstance(callback, LightningModule):
            self._model_callbacks.append(callback)
        elif isinstance(callback, LightingCallback):
            self._lightning_callbacks.append(callback)
        else:
            raise ValueError(f"Unknown callback type: {callback}")

        self._seen_callbacks.add(callback)

    def _move_all_callbacks(self):
        for callback in self._callbacks:
            self._move_callback(callback)

    def setup(
        self,
        module: "Module",
        *optimizers: "Optimizer",
        move_to_device: bool = True,
        _reapply_compile: bool = True,
    ) -> Any:
        output = super().setup(
            module,
            *optimizers,
            move_to_device=move_to_device,
            _reapply_compile=_reapply_compile,
        )

        self._move_all_callbacks()
        return output

    def _call(self, callback_list: list, hook_name: str, *args, **kwargs) -> None:
        # this is the implementation of lightning.Fabric.call
        # but allowing an input for the callback list
        for callback in callback_list:
            method = getattr(callback, hook_name, None)
            if method is None:
                continue

            if not callable(
                method
            ):  # pragma: no cover <- this is using lightning methods that should already be tested
                rank_zero_warn(
                    f"Callback {callback} does not have a callable method {hook_name}"
                )
                continue

            method(*args, **kwargs)

    def call(self, hook_name: str, *args, **kwargs) -> None:
        self._call(self._lcv_callbacks, hook_name, *args, **kwargs)

        # TODO: need to handle these callbacks differently since might have diff names, etc
        # and don't take a CVTrainer
        # self._call(self._model_callbacks, hook_name, *args, **kwargs)
        # self._call(self._lightning_callbacks, hook_name, *args, **kwargs)
        if self._model_callbacks and not self._already_warned_about_callbacks:
            rank_zero_warn(
                "Model callbacks are not yet supported in Fabric. "
                "Please use Lightning callbacks instead."
            )
            self._already_warned_about_callbacks = True

        if self._lightning_callbacks and not self._already_warned_about_callbacks:
            rank_zero_warn(
                "Lightning callbacks are not yet supported in Fabric. "
                "Please use Lightning callbacks instead."
            )
            self._already_warned_about_callbacks = True



import sys
from typing import Any, Literal, Optional, Union
from warnings import warn

import attrs

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    import tomllib as tomli
else:  # pragma: <3.11 cover
    import tomli

from lightning_cv.tuning.utils import FrozenDict
from lightning_cv.typehints import FilePath
from lightning_cv.utils.dataclasses import AttrsDataclassUtilitiesMixin

SuggestValues = Literal["int", "float", "categorical"]


@attrs.define
class Tunable(AttrsDataclassUtilitiesMixin):
    name: str
    suggest: SuggestValues

    def to_optuna_kwargs(self):
        # only include fields that are passed to optuna.Trial.suggest_*
        exclude = {"parent", "map", "suggest"}
        return self.to_dict(exclude=exclude)


@attrs.define
class TunableInt(Tunable):
    suggest: Literal["int"]
    low: int
    high: int
    step: int = 1
    parent: Optional[str] = None


@attrs.define
class TunableFloat(Tunable):
    suggest: Literal["float"]
    low: float
    high: float
    step: Optional[float] = None
    log: bool = False
    parent: Optional[str] = None


@attrs.define
class TunableCategorical(Tunable):
    suggest: Literal["categorical"]
    choices: list
    map: Optional[dict] = None
    parent: Optional[str] = None


TunableType = Union[TunableInt, TunableFloat, TunableCategorical]
MappingT = dict[str, dict[str, dict[str, Any]]]


def _forbidden_converter(value: Optional[list[dict]]) -> Optional[set[FrozenDict]]:
    if value is None:
        return None

    return set(FrozenDict(**combo) for combo in value)


@attrs.define
class HparamConfig(AttrsDataclassUtilitiesMixin):
    hparams: list[TunableType]
    forbidden: Optional[set[FrozenDict]] = attrs.field(
        default=None, converter=_forbidden_converter
    )
    strict: bool = True

    def __attrs_post_init__(self):
        if self.forbidden is not None:
            hparam_names = set(hparam.name for hparam in self.hparams)
            all_forbidden_hparams_subset_of_hparams = all(
                forbidden_hparam_name in hparam_names
                for forbidden in self.forbidden
                for forbidden_hparam_name in forbidden.keys()
            )

            if not all_forbidden_hparams_subset_of_hparams:
                msg = (
                    "Forbidden hparams must be a subset of the hparams defined in the config."
                    f" Received: {self.forbidden}"
                )
                if self.strict:
                    raise ValueError(msg)
                else:
                    msg = f"WARNING: {msg}. This could lead to unexpected behavior if proceeding."
                    warn(msg, RuntimeWarning)

    @classmethod
    def from_file(cls, file: FilePath, strict: bool = True):
        with open(file, "rb") as fp:
            config = tomli.load(fp)

        config["strict"] = config.get("strict", strict)

        return cls.from_dict(config)

    def mappings(self) -> MappingT:
        mapping = {
            hparam.name: hparam.map
            for hparam in self.hparams
            if isinstance(hparam, TunableCategorical) and hparam.map is not None
        }

        return mapping

    def hparams_dict(self):
        return {hparam.name: hparam for hparam in self.hparams}

    def unique_forbidden_combos(self) -> Optional[list[tuple[str, ...]]]:
        if self.forbidden is None:
            return None

        return list(set(tuple(combo.keys()) for combo in self.forbidden))


def get_mappings(file: FilePath) -> MappingT:
    config = HparamConfig.from_file(file)
    return config.mappings()



from typing import TYPE_CHECKING, Any, Mapping, Union

if TYPE_CHECKING:
    from torch import Tensor
    from torch.optim import Optimizer

ConfigOptimizerOutput = Union["Optimizer", Mapping]
ForwardOutputType = Union["Tensor", Mapping[str, Any]]



"""Validators for `attrs.define` dataclass-like config objects."""

from attrs.validators import and_, ge, gt, instance_of, le, lt, or_

is_int = instance_of(int)
is_float = instance_of(float)

positive_int = and_(is_int, gt(0))
non_negative_int = and_(is_int, ge(0))
positive_float = and_(is_float, gt(0.0))
non_negative_float = and_(is_float, ge(0.0))
open_unit_interval = and_(is_float, gt(0.0), lt(1.0))
closed_unit_interval = and_(is_float, ge(0.0), le(1.0))


def equals(expected_value):
    expected_type = type(expected_value)
    return and_(
        instance_of(expected_type),
        ge(expected_value),
        le(expected_value),
    )


_left_open_right_closed_unit_interval = and_(
    instance_of(float),
    gt(0.0),
    le(1.0),
)

limit_batches_validator = or_(
    positive_int,
    _left_open_right_closed_unit_interval,  # type: ignore
)



from collections.abc import Mapping
from dataclasses import dataclass, field, fields
from functools import partial
from typing import Any, Union

from boltons.iterutils import get_path, remap

from lightning_cv.typehints import KwargType


class FrozenDict(Mapping):
    def __init__(self, *args, **kwargs):
        self._d = dict(*args, **kwargs)
        self._hash = None

    def __getitem__(self, key):
        return self._d[key]

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __hash__(self):
        if self._hash is None:  # pragma: no cover <- this is covered
            h = 0
            for pair in self.items():
                h ^= hash(pair)
            self._hash = h
        return self._hash

    def __repr__(self) -> str:
        dict_repr = repr(self._d)
        return f"FrozenDict({dict_repr})"


@dataclass # TODO: switch to attrs?
class _TrialUpdater:
    model: KwargType = field(default_factory=dict)
    data: KwargType = field(default_factory=dict)
    _updated: bool = field(default=False, init=False)

    @staticmethod
    def _update(
        parents: tuple, key: Any, value: Any, update_values: KwargType
    ) -> Union[bool, tuple]:
        """Tries to update leaf nodes based on an arbitrarily nested update schema.

        Almost fits the `boltons.iterutils` `visit` API. It is recommended to create a
        `functools.partial` object with this function.

        Args:
            parents (tuple): parent nodes to the current node
            key (Any): current node
            value (Any): value at current node
            update_values (KwargType): arbitrarily nested update mapping

        Returns:
            bool | tuple: Returns True if at an intermediate node. Otherwise, returns the
                new (node, value) map. See `boltons.iterutils.remap` for more info.
        """

        path = (*parents, key)
        # try to get a value at the current path in update_values
        # but default to the current value
        new_value = (
            # handles nesting
            get_path(root=update_values, path=path, default=None)
            # handles flat
            or update_values.get(key, None)
            # default
            or value
        )

        # skip updating intermediate nodes
        if isinstance(new_value, Mapping):  # pragma: no cover
            return True
        # otherwise update leaf with new value
        return key, new_value

    def update(self, values: KwargType):
        """Update all fields with new values.

        The `values` argument can have a variety of shapes:

        Example:
        ```python
        >>> model_cfg = {
        ...     "in_dim": 64,
        ...     "out_dim": 32,
        ...     "optimizer": {"lr": 1e-3}
        ... }
        >>> data_cfg = {"batch_size": 64}

        # now we want to try a new learning rate and a new batch size
        # the following two `values` schemas work:

        # 1. flat values
        >>> config = _TrialUpdater(model=model_cfg, data=data_cfg)
        >>> new_values = {"lr": 2.53e-4, "batch_size": 16}
        >>> config.update(new_values)
        >>> print(config.model)
        {"in_dim": 64, "out_dim": 32, "optimizer": {"lr": 2.53e-4}}
        >>> print(config.data)
        {"batch_size": 16}

        # 2. nested values
        >>> config = _TrialUpdater(model=model_cfg, data=data_cfg)
        >>> new_values = {
        ...     "model": {"optimizer": {"lr": 2.53e-4}},
        ...     "data": {"batch_size": 16}
        ... }
        >>> config.update(new_values)
        >>> print(config.model)
        {"in_dim": 64, "out_dim": 32, "optimizer": {"lr": 2.53e-4}}
        >>> print(config.data)
        {"batch_size": 16}

        # Further, a combination of flat and nested values also works.
        ```

        Args:
            values (KwargType): an arbitrarily nested map from subfield keys to new values
        """
        for key in fields(self):
            key = key.name
            if key.startswith("_"):
                continue

            current_values: KwargType = getattr(self, key)

            # handles nested and flat update value schemas
            update_values = values.get(key, values)
            update_fn = partial(_TrialUpdater._update, update_values=update_values)
            new_values = remap(root=current_values, visit=update_fn)

            current_values.update(new_values)
        self._updated = True

    def paths(self) -> dict[str, tuple[str, ...]]:
        """Returns a mapping from field names to their paths in the nested structure."""
        paths: dict[str, tuple[str, ...]] = {
            "model": (),
            "data": (),
        }

        stack: list[tuple[KwargType, tuple[str, ...]]] = [
            (self.model, ("model",)),
            (self.data, ("data",)),
        ]

        while stack:
            curr_dict, parents = stack.pop()

            for key, value in curr_dict.items():
                if isinstance(value, dict):
                    stack.append((value, (*parents, key)))

                paths[key] = parents

        return paths

    def to_dict(self) -> dict[str, KwargType]:
        return {
            "model": self.model,
            "data": self.data,
        }



from collections import defaultdict
from typing import Mapping, Union

import optuna

from lightning_cv.tuning.config import HparamConfig, TunableCategorical, TunableType
from lightning_cv.tuning.utils import FrozenDict
from lightning_cv.typehints import FilePath

TrialedValue = Union[int, float, str]
IntFloatStrDict = dict[str, Union[TrialedValue, dict[str, TrialedValue]]]


class HparamRegistry:
    def __init__(self, config: HparamConfig):
        self.config = config
        self.named_config = config.hparams_dict()
        self.uniq_forbidden_combos = config.unique_forbidden_combos()

    @classmethod
    def from_file(cls, config_file: FilePath):
        config = HparamConfig.from_file(config_file)
        return cls(config)

    def _register(self, x: TunableType, trial: optuna.Trial, resample: bool = False):
        method = f"suggest_{x.suggest}"
        method_kwargs = x.to_optuna_kwargs()

        # for resampling trial value, such as for hparam combinations
        # that are forbidden
        if x.name in trial.distributions and resample:
            # need to delete previously sampled value
            # otherwise this will just keep returning the same value
            del trial._cached_frozen_trial.distributions[x.name]

        # no need to return since this will be part of trial.params
        getattr(trial, method)(**method_kwargs)

    def _resolve_forbidden_combinations(self, trial: optuna.Trial):
        if self.config.forbidden is None or self.uniq_forbidden_combos is None:
            return

        hparams = trial.params
        for keys in self.uniq_forbidden_combos:
            current = FrozenDict({key: hparams[key] for key in keys})

            while current in self.config.forbidden:
                new: IntFloatStrDict = dict()
                for key in keys:
                    tunable_type = self.named_config[key]
                    self._register(tunable_type, trial, resample=True)
                    new[key] = trial.params[key]
                current = FrozenDict(new)

    def _nest_hparams(self, trial: optuna.Trial) -> IntFloatStrDict:
        nested_hparams = defaultdict(dict)
        # proxy hparams may map to an arbitrary number of other hparams that are dicts
        # ex: ProxyHparam has choices [a, b, c]
        # but each choice just maps -> a: {real_hparam1: x, real_hparam2: y}
        # thus, if anything mapped, unnest and remove proxy hparam

        # further, need to convert to the correct nesting for the model config
        for key, value in trial.params.items():
            tunable_type = self.named_config[key]
            parent = tunable_type.parent

            if parent is not None:
                # create a new subdict with the parent as the key
                current: dict = nested_hparams[parent]
            else:
                # otherwise add keys to root level
                current = nested_hparams

            if (
                isinstance(tunable_type, TunableCategorical)
                and tunable_type.map is not None
            ):
                value = tunable_type.map[value]

            if isinstance(value, Mapping):
                current.update(value)
            else:
                current[key] = value

        # TODO: we should note that there is a special key called `data`
        # reserved for datamodule/dataloader/data processing hparams
        # nothing really needs to change other than documentation
        return dict(nested_hparams)

    def register_hparams(self, trial: optuna.Trial):
        # get a suggest value for each hparam once
        for hparam in self.config.hparams:
            self._register(hparam, trial, False)

        # then check if any forbidden combinations exist, which need
        # to be resampled
        self._resolve_forbidden_combinations(trial)

        # then we need to build this into the correct nesting
        # based on the parent field for each hparam
        return self._nest_hparams(trial)


def suggest(trial: optuna.Trial, registry: HparamRegistry) -> IntFloatStrDict:
    return registry.register_hparams(trial)



from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping, Optional, TypeVar, Union

if TYPE_CHECKING:
    from lightning.fabric.utilities.types import LRScheduler
    from torch import Tensor
    from torch.optim import Optimizer

from lightning_cv.typehints.data import CVDataLoader, Stage
from lightning_cv.typehints.module import ConfigOptimizerOutput

Number = Union[float, int]
KwargType = dict[str, Any]
MetricType = dict[str, "Tensor"]

SchedulerConfigT = Mapping[str, Union["LRScheduler", bool, str, int]]
OptimizerConfigT = tuple["Optimizer", Optional[SchedulerConfigT]]

_T = TypeVar("_T")

FilePath = Union[str, Path]



from abc import ABC, ABCMeta, abstractmethod
from copy import copy
from typing import Optional

from lightning import LightningDataModule
from torch import arange
from torch.utils.data import DataLoader, Dataset, Subset

from lightning_cv.typehints import KwargType
from lightning_cv.typehints.data import CVDataLoader, Stage
from lightning_cv.typehints.split import CrossValidator


class BaseCVDataModule(ABC, metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, *args, **kwargs) -> None: ...

    @abstractmethod
    def train_val_dataloaders(self, **dataloader_kwargs) -> CVDataLoader: ...

    @abstractmethod
    def setup(self, stage: Stage) -> None: ...


class CrossValidationDataModuleMixin(BaseCVDataModule):
    def __init__(
        self,
        dataset: Dataset,
        batch_size: int,
        cross_validator: type[CrossValidator],
        cross_validator_config: Optional[KwargType] = None,
        dataset_size: Optional[int] = None,
        dataloader_type: type[DataLoader] = DataLoader,
        # attr names
        y_attr: Optional[str] = None,
        group_attr: Optional[str] = None,
        **dataloader_kwargs,
    ) -> None:
        self.dataset = dataset

        # this is customizable if dataset can customize what an individual unit of data is

        if dataset_size is None:
            dataset_size = len(dataset)  # type: ignore

        self.dataset_size = dataset_size

        self.dataloader_type = dataloader_type

        # sklearn CV splitters take X, y, and group args to the .split method
        # since we cant know the structure of the data, we can just split this index tensor
        self._dataset_idx = arange(self.dataset_size)

        self._dataset_y_attr = y_attr
        self._dataset_group_attr = group_attr

        self._validate_dataset()

        self.batch_size = batch_size
        self.dataloader_kwargs = dataloader_kwargs
        self._cross_validator = cross_validator
        self._cross_validator_config = (
            cross_validator_config if cross_validator_config is not None else {}
        )
        self.data_manager = self._cross_validator(**self._cross_validator_config)

    def _overwrite_dataloader_kwargs(self, **kwargs) -> KwargType:
        # overwrite the shared/global dataloader kwargs with kwargs specifically input
        # and then return a copy
        return self.dataloader_kwargs | kwargs

    def _split_train_val_kwargs(self, **kwargs) -> tuple[KwargType, KwargType]:
        train_kwargs = self._overwrite_dataloader_kwargs(**kwargs)
        val_kwargs = copy(train_kwargs)
        val_kwargs["shuffle"] = False
        return train_kwargs, val_kwargs

    def train_val_dataloaders(self, **dataloader_kwargs) -> CVDataLoader:
        train_kwargs, val_kwargs = self._split_train_val_kwargs(**dataloader_kwargs)

        try:
            splits = self.data_manager.split()
        except TypeError:
            # probably sklearn style splitters that sometimes take arguments
            X = self._dataset_idx
            if self._dataset_y_attr is not None:  # pragma: no cover <- already tested
                y = getattr(self.dataset, self._dataset_y_attr, None)
            else:
                y = None

            if self._dataset_group_attr is not None:
                groups = getattr(self.dataset, self._dataset_group_attr, None)
            else:  # pragma: no cover <- already tested
                groups = None

            splits = self.data_manager.split(X, y, groups=groups)

        for train_idx, val_idx in splits:
            train_subset = Subset(self.dataset, train_idx.tolist())
            val_subset = Subset(self.dataset, val_idx.tolist())

            train_loader = self.dataloader_type(
                dataset=train_subset,
                batch_size=self.batch_size,
                **train_kwargs,
            )
            val_loader = self.dataloader_type(
                dataset=val_subset,
                batch_size=self.batch_size,
                **val_kwargs,
            )

            yield train_loader, val_loader

    def _validate_dataset(self):
        if self._dataset_y_attr is not None:
            if not hasattr(self.dataset, self._dataset_y_attr):
                raise ValueError(
                    f"Dataset does not have y attribute called {self._dataset_y_attr}. This is "
                    "used for stratified cross validation strategies in which each fold has "
                    "equal representation of the target variable."
                )

        if self._dataset_group_attr is not None:
            if not hasattr(self.dataset, self._dataset_group_attr):
                raise ValueError(
                    f"Dataset does not have group attribute {self._dataset_group_attr}. This is "
                    "used for group cross validation strategies in which each fold is validated "
                    "with non-target group labels."
                )

    def train_dataloader(self, **dataloader_kwargs):
        raise TypeError(
            "A cross validation datamodule shouldn't need a training dataloader"
        )

    def val_dataloader(self, **dataloader_kwargs):
        raise TypeError(
            "A cross validation datamodule shouldn't need a validation dataloader"
        )


class CrossValidationDataModule(CrossValidationDataModuleMixin, LightningDataModule):
    def __init__(
        self,
        dataset: Dataset,
        batch_size: int,
        cross_validator: type[CrossValidator],
        cross_validator_config: KwargType,
        dataset_size: Optional[int] = None,
        dataloader_type: type[DataLoader] = DataLoader,
        # attr names
        y_attr: Optional[str] = None,
        group_attr: Optional[str] = None,
        **dataloader_kwargs,
    ) -> None:
        LightningDataModule.__init__(self)
        CrossValidationDataModuleMixin.__init__(
            self=self,
            dataset=dataset,
            batch_size=batch_size,
            cross_validator=cross_validator,
            cross_validator_config=cross_validator_config,
            dataset_size=dataset_size,
            dataloader_type=dataloader_type,
            y_attr=y_attr,
            group_attr=group_attr,
            **dataloader_kwargs,
        )

    def setup(self, stage: Stage) -> None:
        if stage != "fit":
            raise ValueError(
                f"Stage {stage} is not supported cross validation datamodules"
            )


CV_DATAMODULE_METHOD_NAME = "train_val_dataloaders"


def is_cross_val_datamodule(obj) -> bool:
    return isinstance(obj, LightningDataModule) and hasattr(
        obj, CV_DATAMODULE_METHOD_NAME
    )



from enum import Enum, auto


class StopReasons(Enum):
    TIMEOUT = auto()
    NO_IMPROVEMENT = auto()
    PRUNED_TRIAL = auto()
    LOSS_NAN_OR_INF = auto()
    PERFORMANCE_STALLED = auto()
    NULL = auto()

    def message(self) -> str:
        if self == StopReasons.NULL:
            return "COMPLETE: Training {max_epochs} epochs completed."

        if self == StopReasons.NO_IMPROVEMENT:
            return (
                "EARLY STOP: Training stopped early due to no improvement in {monitor}"
            )

        if self == StopReasons.TIMEOUT:
            return "TIMEOUT: Training stopped early due to running out of time."

        if self == StopReasons.LOSS_NAN_OR_INF:
            return "NAN/INF: Training stopped early due to NaN or Inf loss."

        if self == StopReasons.PERFORMANCE_STALLED:
            return (
                "STALLED: Training stopped early due to performance stalling. {monitor} is not "
                "improving during the monitored stage."
            )

        return (
            "PRUNED: Training stopped early due to bad trial pruning during hyperparameter "
            "tuning."
        )
